"""
Grouping
========

This example shows how to use the `grouping` argument to visualise vertex covers or clustering, including singletons (i.e. groups of one vertex).

The function takes a few different formats for the groups, but the easiest is a dictionary of sets, in which each key is the name of the group and
each value is the set of vertices in that group.

"""

import igraph as ig
import matplotlib.pyplot as plt
import iplotx as ipx

g = ig.Graph.Ring(8)

ipx.plot(
    network=g,
    layout="circle",
    grouping={
        "singleton": {0},
        "doublet": {1, 2},
        "multiplet": {3, 4, 5, 6, 7},
    },
)
plt.gca().set_aspect(1.0)

# %%
# .. note::
#   Group layouts are computed as rounded convex hulls around each group of vertices. If you are interested in contributing an algorithm that computes concave group layouts, please reach out on GitHub!
#
# Groups can be styles in a similar way as all other elements in `iplotx`:

ipx.plot(
    network=g,
    layout="circle",
    grouping={
        "singleton": {0},
        "doublet": {1, 2},
        "multiplet": {3, 4, 5, 6, 7},
    },
    vertex_size=18,
    vertex_labels=True,
    vertex_label_color="white",
    grouping_facecolor=["purple", "deeppink", "tomato"],
    grouping_edgecolor="black",
    grouping_linewidth=2,
    grouping_vertexpadding=21,
    network_zorder=2,
    grouping_zorder=1,
)
plt.gca().set_aspect(1.0)

# %%
# .. warning::
#   `vertexpadding` and `vertex_size` are **approximately** the same units of measurements, but not exactly. If you need a tight padding, try reducing the
#   padding value below the size of vertices until it's as tight a fit as you need, as shown in the code below.
#
# This can also be achieved with two separate calls, the first one to draw the network and the second to draw the grouping.

layout = g.layout("circle")
fig, ax = plt.subplots(figsize=(5, 5))
ipx.plot(
    network=g,
    layout=layout,
    ax=ax,
    vertex_size=18,
    vertex_labels=True,
    vertex_label_color="white",
    zorder=2,
)
ipx.plot(
    grouping={
        "singleton": {0},
        "doublet": {1, 2},
        "multiplet": {3, 4, 5, 6, 7},
    },
    layout=layout,
    ax=ax,
    grouping_facecolor=["purple", "deeppink", "tomato"],
    grouping_edgecolor="black",
    grouping_linewidth=2,
    grouping_vertexpadding=12,  # Note the tighter padding
    zorder=1,
)

# %%
# .. note::
#   The argument `zorder` is applied to both network and grouping style, however since only one of them is plotted each
#   time this is equivalent to setting `network_zorder` and `grouping_zorder`. If both are specified, `zorder` takes
#   priority over its more specific cousins.



"""
igraph/networkx compatibility
=============================

The same graph in igraph and networkx, plotted with iplotx.
"""

import igraph as ig
import networkx as nx
import matplotlib.pyplot as plt
import iplotx as ipx

fig, axs = plt.subplots(1, 2, figsize=(6, 3))

axs[0].set_title("igraph")
g1 = ig.Graph.Ring(5, directed=True)
layout1 = g1.layout("circle")
ipx.plot(g1, ax=axs[0], layout=layout1)

axs[1].set_title("networkx")
g2 = nx.cycle_graph(n=5, create_using=nx.DiGraph)
layout2 = nx.layout.circular_layout(g2)
ipx.plot(g2, ax=axs[1], layout=layout2)



"""
Typing module for data/object ingestion. This module described the abstract data types that providers need to comply with to be compatible with iplotx.

Networkx and trees are treated separately for practical reasons: many tree analysis libraries rely heavily on recursive data structures, which do not
work as well on general networks.
"""

from typing import (
    NotRequired,
    TypedDict,
    Protocol,
    Optional,
    Sequence,
    Any,
    Iterable,
)
from collections.abc import Hashable
import numpy as np
import pandas as pd
from ..typing import (
    GraphType,
    LayoutType,
    TreeType,
)
from .heuristics import (
    normalise_tree_layout,
)


class NetworkData(TypedDict):
    """Network data structure for iplotx."""

    directed: bool
    vertex_df: pd.DataFrame
    edge_df: pd.DataFrame
    ndim: int
    network_library: NotRequired[str]


class NetworkDataProvider(Protocol):
    """Protocol for network data ingestion provider for iplotx."""

    def __call__(
        self,
        network: GraphType,
        layout: Optional[LayoutType] = None,
        vertex_labels: Optional[Sequence[str] | dict[Hashable, str] | pd.Series] = None,
        edge_labels: Optional[Sequence[str] | dict] = None,
    ) -> NetworkData:
        """Create network data object for iplotx from any provider."""
        raise NotImplementedError("Network data providers must implement this method.")

    @staticmethod
    def check_dependencies():
        """Check whether the dependencies for this provider are installed."""
        raise NotImplementedError("Network data providers must implement this method.")

    @staticmethod
    def graph_type():
        """Return the graph type from this provider to check for instances."""
        raise NotImplementedError("Network data providers must implement this method.")


class TreeData(TypedDict):
    """Tree data structure for iplotx."""

    rooted: bool
    directed: bool | str
    root: Optional[Hashable]
    leaf_df: pd.DataFrame
    vertex_df: dict[Hashable, tuple[float, float]]
    edge_df: dict[Hashable, Sequence[tuple[float, float]]]
    layout_coordinate_system: str
    layout_name: str
    orientation: str
    ndim: int
    tree_library: NotRequired[str]


class TreeDataProvider(Protocol):
    """Protocol for tree data ingestion provider for iplotx."""

    def __init__(
        self,
        tree: TreeType,
    ) -> None:
        """Initialize the provider with the tree type.

        Parameters:
            tree: The tree type that this provider will handle.
        """
        self.tree = tree

    @staticmethod
    def check_dependencies():
        """Check whether the dependencies for this provider are installed."""
        raise NotImplementedError("Tree data providers must implement this method.")

    @staticmethod
    def tree_type():
        """Return the tree type from this provider to check for instances."""
        raise NotImplementedError("Tree data providers must implement this method.")

    def is_rooted(self) -> bool:
        """Get whether the tree is rooted.

        Returns:
            A boolean indicating whether the tree is rooted.

        Note: This is a default implemntation that can be overridden by the provider
        if they support unrooted trees (e.g. Biopython).
        """
        return True

    def get_root(self) -> Any:
        """Get the tree root in a provider-specific data structure.

        Returns:
            The root of the tree.

        Note: This is a default implemntation that can be overridden by the provider.
        """
        root_attr = self.tree.root
        if callable(root_attr):
            return root_attr()
        else:
            return root_attr

    def get_leaves(self) -> Sequence[Any]:
        """Get the tree leaves/tips in a provider-specific data structure.

        Returns:
            The leaves or tips of the tree.
        """
        raise NotImplementedError("Tree data providers must implement this method.")

    def preorder(self) -> Iterable[Any]:
        """Preorder (DFS - parent first) iteration over the tree.

        Returns:
            An iterable of nodes in preorder traversal.
        """
        raise NotImplementedError("Tree data providers must implement this method.")

    def postorder(self) -> Iterable[Any]:
        """Postorder (DFS - child first) iteration over the tree.

        Returns:
            An iterable of nodes in preorder traversal.
        """
        raise NotImplementedError("Tree data providers must implement this method.")

    @staticmethod
    def get_children(
        node: Any,
    ) -> Sequence[Any]:
        """Get the children of a node.

        Parameters:
            node: The node to get the children from.
        Returns:
            A sequence of children nodes.
        """
        raise NotImplementedError("Tree data providers must implement this method.")

    @staticmethod
    def get_branch_length(
        node: Any,
    ) -> Optional[float]:
        """Get the length of the branch to this node.

        Parameters:
            node: The node to get the branch length from.
        Returns:
            The branch length to the node.
        """
        raise NotImplementedError("Tree data providers must implement this method.")

    def get_branch_length_default_to_one(
        self,
        node: Any,
    ) -> float:
        """Get the length of the branch to this node, defaulting to 1.0 if not available.

        Parameters:
            node: The node to get the branch length from.
        Returns:
            The branch length to the node, defaulting to 1.0 if not available.
        """
        branch_length = self.get_branch_length(node)
        return branch_length if branch_length is not None else 1.0

    def __call__(
        self,
        layout: str | LayoutType,
        orientation: Optional[str],
        layout_style: Optional[dict[str, int | float | str]] = None,
        directed: bool | str = False,
        vertex_labels: Optional[
            Sequence[str] | dict[Hashable, str] | pd.Series | bool
        ] = None,
        edge_labels: Optional[Sequence[str] | dict] = None,
        leaf_labels: Optional[Sequence[str] | dict[Hashable, str] | pd.Series] = None,
    ) -> TreeData:
        """Create tree data object for iplotx from ete4.core.tre.Tree classes."""

        if layout_style is None:
            layout_style = {}

        if orientation is None:
            if layout == "horizontal":
                orientation = "right"
            elif layout == "vertical":
                orientation = "descending"
            elif layout == "radial":
                orientation = "clockwise"

        tree_data = {
            "root": self.get_root(),
            "rooted": self.is_rooted(),
            "directed": directed,
            "ndim": 2,
            "layout_name": layout,
            "orientation": orientation,
        }

        # Add vertex_df including layout
        tree_data["vertex_df"] = normalise_tree_layout(
            layout,
            orientation=orientation,
            root=tree_data["root"],
            preorder_fun=self.preorder,
            postorder_fun=self.postorder,
            children_fun=self.get_children,
            branch_length_fun=self.get_branch_length_default_to_one,
            **layout_style,
        )
        if layout in ("radial",):
            tree_data["layout_coordinate_system"] = "polar"
        else:
            tree_data["layout_coordinate_system"] = "cartesian"

        # Add edge_df
        edge_data = {"_ipx_source": [], "_ipx_target": []}
        for node in self.preorder():
            for child in self.get_children(node):
                if directed == "parent":
                    edge_data["_ipx_source"].append(child)
                    edge_data["_ipx_target"].append(node)
                else:
                    edge_data["_ipx_source"].append(node)
                    edge_data["_ipx_target"].append(child)
        edge_df = pd.DataFrame(edge_data)
        tree_data["edge_df"] = edge_df

        # Add leaf_df
        tree_data["leaf_df"] = pd.DataFrame(index=self.get_leaves())

        # Add vertex labels
        if vertex_labels is None:
            vertex_labels = False
        if np.isscalar(vertex_labels) and vertex_labels:
            tree_data["vertex_df"]["label"] = [
                x.name for x in tree_data["vertex_df"].index
            ]
        elif not np.isscalar(vertex_labels):
            # If a dict-like object is passed, it can be incomplete (e.g. only the leaves):
            # we fill the rest with empty strings which are not going to show up in the plot.
            if isinstance(vertex_labels, pd.Series):
                vertex_labels = dict(vertex_labels)
            if isinstance(vertex_labels, dict):
                for vertex in tree_data["vertex_df"].index:
                    if vertex not in vertex_labels:
                        vertex_labels[vertex] = ""
            tree_data["vertex_df"]["label"] = pd.Series(vertex_labels)

        # Add leaf labels
        if leaf_labels is None:
            leaf_labels = False
        if np.isscalar(leaf_labels) and leaf_labels:
            tree_data["leaf_labels"]["label"] = [
                # FIXME: this is likely broken
                x.name
                for x in tree_data["leaf_df"].index
            ]
        elif not np.isscalar(leaf_labels):
            # Leaves are already in the dataframe in a certain order, so sequences are allowed
            if isinstance(leaf_labels, (list, tuple, np.ndarray)):
                leaf_labels = {
                    leaf: label
                    for leaf, label in zip(tree_data["leaf_df"].index, leaf_labels)
                }
            # If a dict-like object is passed, it can be incomplete (e.g. only the leaves):
            # we fill the rest with empty strings which are not going to show up in the plot.
            if isinstance(leaf_labels, pd.Series):
                leaf_labels = dict(leaf_labels)
            if isinstance(leaf_labels, dict):
                for leaf in tree_data["leaf_df"].index:
                    if leaf not in leaf_labels:
                        leaf_labels[leaf] = ""
            tree_data["leaf_df"]["label"] = pd.Series(leaf_labels)

        return tree_data



"""
Utility function for tests.

Mostly functions adapted from matplotlib.testing.

NOTE: folders for expected and generated images are as explained at:
https://github.com/matplotlib/matplotlib/blob/def8fa4eaf6b97958b0a6b48a4da2558ba0942c4/lib/matplotlib/testing/decorators.py#L457
i.e.:
   For test module ``foo.bar.test_baz``, the baseline directory is at
   ``foo/bar/baseline_images/test_baz`` and the result directory at
   ``$(pwd)/result_images/test_baz``.  The result directory is created if it
   doesn't exist.
"""

import sys
import inspect
import functools

try:
    import matplotlib
    from matplotlib.testing.decorators import _collect_new_figures, _ImageComparisonBase
    import matplotlib.pyplot as plt
except ImportError:
    matplotlib = None


# NOTE: Parametrizing this requires pytest (see matplotlib's test suite)
_default_extension = "png"


def _unittest_image_comparison(
    baseline_images, tol, remove_text, savefig_kwargs, style
):
    """
    Decorate function with image comparison for unittest.
    This function creates a decorator that wraps a figure-generating function
    with image comparison code.
    """

    def decorator(func):
        old_sig = inspect.signature(func)

        @functools.wraps(func)
        @matplotlib.style.context(style)
        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            __tracebackhide__ = True

            img = _ImageComparisonBase(
                func, tol=tol, remove_text=remove_text, savefig_kwargs=savefig_kwargs
            )
            matplotlib.testing.set_font_settings_for_testing()

            with _collect_new_figures() as figs:
                plt.close("all")
                func(*args, **kwargs)

            assert len(figs) == len(
                baseline_images
            ), "Test generated {} images but there are {} baseline images".format(
                len(figs), len(baseline_images)
            )
            for fig, baseline in zip(figs, baseline_images):
                img.compare(fig, baseline, _default_extension, _lock=False)

        parameters = list(old_sig.parameters.values())
        new_sig = old_sig.replace(parameters=parameters)
        wrapper.__signature__ = new_sig

        return wrapper

    return decorator


def image_comparison(
    baseline_images,
    tol=0.025,
    remove_text=False,
    savefig_kwarg=None,
    # Default of mpl_test_settings fixture and cleanup too.
    style=("classic", "_classic_test_patch"),
):
    """
    Compare images generated by the test with those specified in
    *baseline_images*, which must correspond, else an `ImageComparisonFailure`
    exception will be raised.
    Parameters
    ----------
    baseline_images : list or None
        A list of strings specifying the names of the images generated by
        calls to `.Figure.savefig`.
        If *None*, the test function must use the ``baseline_images`` fixture,
        either as a parameter or with `pytest.mark.usefixtures`. This value is
        only allowed when using pytest.
    tol : float, default: 0
        The RMS threshold above which the test is considered failed.
        Due to expected small differences in floating-point calculations, on
        32-bit systems an additional 0.06 is added to this threshold.
    remove_text : bool
        Remove the title and tick text from the figure before comparison.  This
        is useful to make the baseline images independent of variations in text
        rendering between different versions of FreeType.
        This does not remove other, more deliberate, text, such as legends and
        annotations.
    savefig_kwarg : dict
        Optional arguments that are passed to the savefig method.
    style : str, dict, or list
        The optional style(s) to apply to the image test. The test itself
        can also apply additional styles if desired. Defaults to ``["classic",
        "_classic_test_patch"]``.
    """

    if savefig_kwarg is None:
        savefig_kwarg = {}  # default no kwargs to savefig
    if sys.maxsize <= 2**32:
        tol += 0.06
    return _unittest_image_comparison(
        baseline_images=baseline_images,
        tol=tol,
        remove_text=remove_text,
        savefig_kwargs=savefig_kwarg,
        style=style,
    )



from typing import (
    Any,
    Optional,
    Sequence,
)
from functools import partialmethod

from ...typing import (
    TreeDataProvider,
)


class BiopythonDataProvider(TreeDataProvider):
    def is_rooted(self) -> bool:
        return self.tree.rooted

    def _traverse(self, order: str) -> Any:
        """Traverse the tree."""
        return self.tree.find_clades(order=order)

    preorder = partialmethod(_traverse, order="preorder")
    postorder = partialmethod(_traverse, order="postorder")

    def get_leaves(self) -> Sequence[Any]:
        return self.tree.get_terminals()

    @staticmethod
    def get_children(node: Any) -> Sequence[Any]:
        return node.clades

    @staticmethod
    def get_branch_length(node: Any) -> Optional[float]:
        return node.branch_length

    @staticmethod
    def check_dependencies() -> bool:
        try:
            from Bio import Phylo
        except ImportError:
            return False
        return True

    @staticmethod
    def tree_type():
        from Bio import Phylo

        return Phylo.BaseTree.Tree



from typing import Optional, Sequence
import numpy as np
import pandas as pd
import matplotlib as mpl

from .typing import (
    GraphType,
    LayoutType,
)
from .style import (
    get_style,
    rotate_style,
)
from .utils.matplotlib import (
    _stale_wrapper,
    _forwarder,
    _build_cmap_fun,
)
from .ingest import (
    ingest_network_data,
)
from .vertex import (
    VertexCollection,
)
from .edge import (
    EdgeCollection,
    make_stub_patch as make_undirected_edge_patch,
)


@_forwarder(
    (
        "set_clip_path",
        "set_clip_box",
        "set_snap",
        "set_sketch_params",
        "set_animated",
        "set_picker",
    )
)
class NetworkArtist(mpl.artist.Artist):
    def __init__(
        self,
        network: GraphType,
        layout: Optional[LayoutType] = None,
        vertex_labels: Optional[list | dict | pd.Series] = None,
        edge_labels: Optional[Sequence] = None,
        transform: mpl.transforms.Transform = mpl.transforms.IdentityTransform(),
        offset_transform: Optional[mpl.transforms.Transform] = None,
    ):
        """Network container artist that groups all plotting elements.

        Parameters:
            network: The network to plot.
            layout: The layout of the network. If None, this function will attempt to
                infer the layout from the network metadata, using heuristics. If that fails, an
                exception will be raised.
            vertex_labels: The labels for the vertices. If None, no vertex labels
                will be drawn. If a list, the labels are taken from the list. If a dict, the keys
                should be the vertex IDs and the values should be the labels.
            edge_labels: The labels for the edges. If None, no edge labels will be drawn.

        """
        self.network = network
        self._ipx_internal_data = ingest_network_data(
            network,
            layout,
            vertex_labels=vertex_labels,
            edge_labels=edge_labels,
        )

        super().__init__()

        # This is usually the identity (which scales poorly with dpi)
        self.set_transform(transform)

        # This is usually transData
        self.set_offset_transform(offset_transform)

        zorder = get_style(".network").get("zorder", 1)
        self.set_zorder(zorder)

        self._add_vertices()
        self._add_edges()

    def get_children(self):
        return (self._vertices, self._edges)

    def set_figure(self, fig):
        super().set_figure(fig)
        for child in self.get_children():
            child.set_figure(fig)

    def get_offset_transform(self):
        """Get the offset transform (for vertices/edges)."""
        return self._offset_transform

    def set_offset_transform(self, offset_transform):
        """Set the offset transform (for vertices/edges)."""
        self._offset_transform = offset_transform

    def get_vertices(self):
        """Get VertexCollection artist."""
        return self._vertices

    def get_edges(self):
        """Get EdgeCollection artist."""
        return self._edges

    def get_vertex_labels(self):
        """Get list of vertex label artists."""
        return self._vertices.get_labels()

    def get_edge_labels(self):
        """Get list of edge label artists."""
        return self._edges.get_labels()

    def get_datalim(self, transData, pad=0.15):
        """Get limits on x/y axes based on the graph layout data.

        Parameters:
            transData (Transform): The transform to use for the data.
            pad (float): Padding to add to the limits. Default is 0.05.
                Units are a fraction of total axis range before padding.
        """
        layout = self.get_layout().values

        if len(layout) == 0:
            return mpl.transforms.Bbox([[0, 0], [1, 1]])

        bbox = mpl.transforms.Bbox.union(
            [
                self._vertices.get_datalim(transData),
                self._edges.get_datalim(transData),
            ]
        )

        bbox = bbox.expanded(sw=(1.0 + pad), sh=(1.0 + pad))
        return bbox

    def autoscale_view(self, tight=False):
        """Recompute data limits from this artist and set autoscale based on them."""
        bbox = self.get_datalim(self.axes.transData)
        self.axes.update_datalim(bbox)
        self.axes.autoscale_view(tight=tight)

    def get_layout(self):
        layout_columns = [
            f"_ipx_layout_{i}" for i in range(self._ipx_internal_data["ndim"])
        ]
        vertex_layout_df = self._ipx_internal_data["vertex_df"][layout_columns]
        return vertex_layout_df

    def _get_label_series(self, kind):
        if "label" in self._ipx_internal_data[f"{kind}_df"].columns:
            return self._ipx_internal_data[f"{kind}_df"]["label"]
        else:
            return None

    def _add_vertices(self):
        """Add vertices to the network artist."""

        self._vertices = VertexCollection(
            layout=self.get_layout(),
            layout_coordinate_system=self._ipx_internal_data.get(
                "layout_coordinate_system", "cartesian"
            ),
            style=get_style(".vertex"),
            labels=self._get_label_series("vertex"),
            transform=self.get_transform(),
            offset_transform=self.get_offset_transform(),
        )

    def _add_edges(self):
        """Add edges to the network artist.

        NOTE: UndirectedEdgeCollection and ArrowCollection are both subclasses of
        PatchCollection. When used with a cmap/norm, they set their facecolor
        according to the cmap, even though most likely we only want the edgecolor
        set that way. It can make for funny looking plots that are not uninteresting
        but mostly niche at this stage. Therefore we sidestep the whole cmap thing
        here.
        """

        labels = self._get_label_series("edge")
        edge_style = get_style(".edge")

        if "cmap" in edge_style:
            cmap_fun = _build_cmap_fun(
                edge_style["color"],
                edge_style["cmap"],
            )
        else:
            cmap_fun = None

        edge_df = self._ipx_internal_data["edge_df"].set_index(
            ["_ipx_source", "_ipx_target"]
        )

        if "cmap" in edge_style:
            colorarray = []
        edgepatches = []
        adjacent_vertex_ids = []
        for i, (vid1, vid2) in enumerate(edge_df.index):
            edge_stylei = rotate_style(edge_style, index=i, key=(vid1, vid2))

            # FIXME:: Improve this logic. We have three layers of priority:
            # 1. Explicitely set in the style of "plot"
            # 2. Internal through network attributes
            # 3. Default styles
            # Because 1 and 3 are merged as a style context on the way in,
            # it's hard to squeeze 2 in the middle. For now, we will assume
            # the priority order is 2-1-3 instead (internal property is
            # highest priority).
            # This is also why we cannot shift this logic further into the
            # EdgeCollection class, which is oblivious of NetworkArtist's
            # internal data. In fact, one would argue this needs to be
            # pushed outwards to deal with the wrong ordering.
            _update_from_internal(edge_stylei, edge_df.iloc[i], kind="edge")

            if cmap_fun is not None:
                colorarray.append(edge_stylei["color"])
                edge_stylei["color"] = cmap_fun(edge_stylei["color"])

            # These are not the actual edges drawn, only stubs to establish
            # the styles which are then fed into the dynamic, optimised
            # factory (the collection) below
            patch = make_undirected_edge_patch(
                **edge_stylei,
            )
            edgepatches.append(patch)
            adjacent_vertex_ids.append((vid1, vid2))

        if "cmap" in edge_style:
            vmin = np.min(colorarray)
            vmax = np.max(colorarray)
            norm = mpl.colors.Normalize(vmin=vmin, vmax=vmax)
            edge_style["norm"] = norm

        self._edges = EdgeCollection(
            edgepatches,
            vertex_ids=adjacent_vertex_ids,
            vertex_collection=self._vertices,
            labels=labels,
            transform=self.get_offset_transform(),
            style=edge_style,
            directed=self._ipx_internal_data["directed"],
        )
        if "cmap" in edge_style:
            self._edges.set_array(colorarray)

    @_stale_wrapper
    def draw(self, renderer):
        """Draw each of the children, with some buffering mechanism."""
        if not self.get_children():
            self._add_vertices()
            self._add_edges()

        if not self.get_visible():
            return

        # FIXME: Callbacks on stale vertices/edges??

        # NOTE: looks like we have to manage the zorder ourselves
        # this is kind of funny actually
        children = list(self.get_children())
        children.sort(key=lambda x: x.zorder)
        for art in children:
            art.draw(renderer)


def _update_from_internal(style, row, kind):
    """Update single vertex/edge style from internal data."""
    if "color" in row:
        style["color"] = row["color"]
    if "facecolor" in row:
        style["facecolor"] = row["facecolor"]
    if "edgecolor" in row:
        if kind == "vertex":
            style["edgecolor"] = row["edgecolor"]
        else:
            style["color"] = row["edgecolor"]

    if "linewidth" in row:
        style["linewidth"] = row["linewidth"]
    if "linestyle" in row:
        style["linestyle"] = row["linestyle"]
    if "alpha" in row:
        style["alpha"] = row["alpha"]



"""
Labels and colors
=================

This example from networkx demonstrates how to use transparency to overlay multiple plots with distinct highlights.
"""

import networkx as nx
import matplotlib.pyplot as plt
import iplotx as ipx

G = nx.cubical_graph()
pos = nx.spring_layout(G, seed=3113794652)  # positions for all nodes
nx.set_node_attributes(G, pos, "pos")  # Will not be needed after PR 7571
labels = iter(
    [
        r"$a$",
        r"$b$",
        r"$c$",
        r"$d$",
        r"$\alpha$",
        r"$\beta$",
        r"$\gamma$",
        r"$\delta$",
    ]
)
nx.set_node_attributes(
    G,
    {
        n: {
            "color": "tab:red" if n < 4 else "tab:blue",
            "label": next(labels),
        }
        for n in G.nodes()
    },
)

fig, ax = plt.subplots(figsize=(5, 4))
ipx.plot(
    G,
    ax=ax,
    layout="pos",
    vertex_labels=True,
    style={
        "vertex": {
            "facecolor": G.nodes.data("color"),
            "alpha": G.nodes.data("alpha"),
            "size": 40,
            "label": {"size": 22, "color": "whitesmoke"},
            "zorder": 7,
        },
        "edge": {
            "color": "tab:grey",
        },
    },
)
for i in range(2):
    subG = G.subgraph(range(i * 4, (i + 1) * 4))
    ipx.plot(
        subG,
        layout="pos",
        ax=ax,
        style={
            "edge": {
                "color": ["tab:red", "tab:blue"][i],
                "alpha": 0.5,
                "linewidth": 8,
                "zorder": 0,
            },
            "vertex": {
                "size": 35,
                "alpha": 0,
            },
        },
    )
plt.tight_layout()



"""
Simple example from networkx
============================

This example demonstrates how to use `iplotx` with `networkx`. Note how the vertex size is computed dynamically based on label size.
"""

import networkx as nx
import matplotlib.pyplot as plt
import iplotx as ipx

G = nx.Graph()
G.add_edge(1, 2)
G.add_edge(1, 3)
G.add_edge(1, 5)
G.add_edge(2, 3)
G.add_edge(3, 4)
G.add_edge(4, 5)

# explicitly set positions
pos = {1: (0, 0), 2: (-1, 0.3), 3: (2, 0.17), 4: (4, 0.255), 5: (5, 0.03)}

ipx.plot(
    G,
    layout=pos,
    vertex_labels=True,
    style={
        "vertex": {
            "size": "label",
            "facecolor": "white",
            "edgecolor": "black",
            "linewidth": 5,
            "label": {
                "size": 26,
                "color": "black",
            },
        },
        "edge": {
            "linewidth": 5,
        },
    },
)

# Set margins for the axes so that nodes aren't clipped
ax = plt.gca()
ax.margins(0.20)



"""
Layout functions, currently limited to trees.
"""

from typing import Any
from collections.abc import (
    Hashable,
    Callable,
)

import numpy as np


def compute_tree_layout(
    layout: str,
    orientation: str,
    root: Any,
    preorder_fun: Callable,
    postorder_fun: Callable,
    children_fun: Callable,
    branch_length_fun: Callable,
    **kwargs,
) -> dict[Hashable, list[float]]:
    """Compute the layout for a tree.

    Parameters:
        layout: The name of the layout, e.g. "horizontal", "vertial", or "radial".
        orientation: The orientation of the layout, e.g. "right", "left", "descending",
            "ascending", "clockwise", "anticlockwise".

    Returns:
        A layout dictionary with node positions.
    """
    kwargs["root"] = root
    kwargs["preorder_fun"] = preorder_fun
    kwargs["postorder_fun"] = postorder_fun
    kwargs["children_fun"] = children_fun
    kwargs["branch_length_fun"] = branch_length_fun
    kwargs["orientation"] = orientation

    if layout == "radial":
        layout_dict = _radial_tree_layout(**kwargs)
    elif layout == "horizontal":
        layout_dict = _horizontal_tree_layout(**kwargs)
    elif layout == "vertical":
        layout_dict = _vertical_tree_layout(**kwargs)
    else:
        raise ValueError(f"Tree layout not available: {layout}")

    return layout_dict


def _horizontal_tree_layout_right(
    root: Any,
    preorder_fun: Callable,
    postorder_fun: Callable,
    children_fun: Callable,
    branch_length_fun: Callable,
) -> dict[Hashable, list[float]]:
    """Build a tree layout horizontally, left to right.

    The strategy is the usual one:
    1. Compute the y values for the leaves, from 0 upwards.
    2. Compute the y values for the internal nodes, bubbling up (postorder).
    3. Set the x value for the root as 0.
    4. Compute the x value of all nodes, trickling down (BFS/preorder).
    5. Compute the edges from the end nodes.
    """
    layout = {}

    # Set the y values for vertices
    i = 0
    for node in postorder_fun():
        children = children_fun(node)
        if len(children) == 0:
            layout[node] = [None, i]
            i += 1
        else:
            layout[node] = [
                None,
                np.mean([layout[child][1] for child in children]),
            ]

    # Set the x values for vertices
    layout[root][0] = 0
    for node in preorder_fun():
        for child in children_fun(node):
            bl = branch_length_fun(child)
            if bl is None:
                bl = 1.0
            layout[child][0] = layout[node][0] + bl

    return layout


def _horizontal_tree_layout(
    orientation="right",
    **kwargs,
) -> dict[Hashable, list[float]]:
    """Horizontal tree layout."""
    if orientation not in ("right", "left"):
        raise ValueError("Orientation must be 'right' or 'left'.")

    layout = _horizontal_tree_layout_right(**kwargs)

    if orientation == "left":
        for key in layout:
            layout[key][0] *= -1
    return layout


def _vertical_tree_layout(
    orientation="descending",
    **kwargs,
) -> dict[Hashable, list[float]]:
    """Vertical tree layout."""
    sign = -1 if orientation == "descending" else 1
    layout = _horizontal_tree_layout(**kwargs)
    for key, value in layout.items():
        # Invert x and y
        layout[key] = value[::-1]
        # Orient vertically
        layout[key][1] *= sign
    return layout


def _radial_tree_layout(
    orientation: str = "right",
    start: float = 180,
    span: float = 360,
    **kwargs,
) -> dict[Hashable, tuple[float, float]]:
    """Radial tree layout.

    Parameters:
        orientation: Whether the layout fans out towards the right (clockwise) or left
            (anticlockwise).
        start: The starting angle in degrees, default is -180 (left).
        span: The angular span in degrees, default is 360 (full circle). When this is
            360, it leaves a small gap at the end to ensure the first and last leaf
            are not overlapping.
    Returns:
        A dictionary with the radial layout.
    """
    # Short form
    th = start * np.pi / 180
    th_span = span * np.pi / 180
    pad = int(span == 360)
    sign = -1 if orientation in ("right", "clockwise") else 1

    layout = _horizontal_tree_layout_right(**kwargs)
    ymax = max(point[1] for point in layout.values())
    for key, (x, y) in layout.items():
        r = x
        theta = sign * th_span * y / (ymax + pad) + th
        # We export r and theta to ensure theta does not
        # modulo 2pi if we take the tan and then arctan later.
        layout[key] = (r, theta)

    return layout



"""
Basics of styling
======================

This example shows a few simple styling options available in `iplotx`.
"""

import igraph as ig
import matplotlib.pyplot as plt
import iplotx as ipx

g = ig.Graph.Ring(5)
layout = g.layout("circle").coords
style = {
    "vertex": {
        "facecolor": ["purple", "pink", "black"],
        "edgecolor": "black",
    },
    "edge": {
        "linestyle": "--",
        "linewidth": 2,
        "color": "deeppink",
    },
}
fig, ax = plt.subplots(figsize=(3, 3))
ipx.plot(g, ax=ax, layout=layout, style=style)

# %%
# There is no question that this network looks much nicer now.



"""
Module for vertex groupings code, especially the GroupingArtist class.
"""

from typing import Union
import numpy as np
import matplotlib as mpl
from matplotlib.collections import PatchCollection


from .typing import (
    GroupingType,
    LayoutType,
)
from .ingest.heuristics import (
    normalise_layout,
    normalise_grouping,
)
from .style import get_style, rotate_style
from .utils.geometry import (
    convex_hull,
    _compute_group_path_with_vertex_padding,
)


class GroupingArtist(PatchCollection):
    """Matplotlib artist for a vertex grouping (clustering/cover).

    This class is used to plot patches surrounding groups of vertices in a network.
    """

    _factor = 1.0

    def __init__(
        self,
        grouping: GroupingType,
        layout: LayoutType,
        vertexpadding: Union[None, int] = None,
        points_per_curve: int = 30,
        transform: mpl.transforms.Transform = mpl.transforms.IdentityTransform(),
        *args,
        **kwargs,
    ) -> None:
        """Container artist for vertex groupings, e.g. covers or clusterings.

        Parameters:
            grouping: This can be a sequence of sets (a la networkx), an igraph Clustering
                or Cover instance (including VertexClustering/VertexCover), or a sequence
                of integers/strings indicating memberships for each vertex.
            layout: The layout of the vertices. If this object has no keys/index, the
                vertices are assumed to have IDs corresponding to integers starting from
                zero.
            vertexpadding: How may points of padding to leave around each vertex centre.
            points_per_curve: How many points to use to approximate a round envelope around
                each convex hull vertex.
            transform: The matplotlib transform to use for the patches (typically transData).
        """
        if vertexpadding is not None:
            self._vertexpadding = vertexpadding
        else:
            style = get_style(".grouping")
            self._vertexpadding = style.get("vertexpadding", 10)

        self._points_per_curve = points_per_curve

        network = kwargs.pop("network", None)
        patches, grouping, coords_hulls = self._create_patches(
            grouping,
            layout,
            network,
            **kwargs,
        )
        if "network" in kwargs:
            del kwargs["network"]
        self._grouping = grouping
        self._coords_hulls = coords_hulls
        kwargs["match_original"] = True

        super().__init__(patches, *args, **kwargs)

        zorder = get_style(".grouping").get("zorder", 1)
        self.set_zorder(zorder)

        self.set_transform(transform)

    def set_figure(self, figure) -> None:
        """Set the figure for the grouping, recomputing the paths depending on the figure's dpi."""
        ret = super().set_figure(figure)
        self._compute_paths(self.get_figure(root=True).dpi)
        return ret

    def get_vertexpadding(self) -> float:
        """Get the vertex padding of each group."""
        return self._vertexpadding

    def get_vertexpadding_dpi(self, dpi: float = 72.0) -> float:
        """Get vertex padding of each group, scaled by dpi of the figure."""
        return self.get_vertexpadding() * dpi / 72.0 * self._factor

    def _create_patches(self, grouping, layout, network, **kwargs):
        layout = normalise_layout(layout, network=network)
        grouping = normalise_grouping(grouping, layout)
        style = get_style(".grouping")
        style.pop("vertexpadding", None)

        style.update(kwargs)

        patches = []
        coords_hulls = []
        for i, (name, vids) in enumerate(grouping.items()):
            if len(vids) == 0:
                continue
            vids = np.array(list(vids))
            coords = layout.loc[vids].values
            idx_hull = convex_hull(coords)
            coords_hull = coords[idx_hull]
            coords_hulls.append(coords_hull)

            stylei = rotate_style(style, i)

            # NOTE: the transform is set later on
            patch = _compute_group_patch_stub(
                coords_hull,
                self._vertexpadding,
                label=name,
                **stylei,
            )

            patches.append(patch)
        return patches, grouping, coords_hulls

    def _compute_paths(self, dpi: float = 72.0) -> None:
        ppc = self._points_per_curve
        for i, hull in enumerate(self._coords_hulls):
            self._paths[i].vertices = _compute_group_path_with_vertex_padding(
                hull,
                self._paths[i].vertices,
                self.get_transform(),
                vertexpadding=self.get_vertexpadding_dpi(dpi),
                points_per_curve=ppc,
            )

    def _process(self) -> None:
        self._compute_paths()

    def draw(self, renderer) -> None:
        """Draw or re-draw the grouping patches.

        Parameters:
            renderer: The renderer to use for drawing the patches.
        """
        # FIXME: this kind of breaks everything since the vertices' magical "_transforms" does
        # not really scale from 72 pixels but rather from the screen's or something.
        # Conclusion: using this keeps consistency across dpis but breaks proportionality of
        # vertexpadding and vertex_size (for now).
        # NOTE: this might be less bad than initially thought in the sense that even perfect
        # scaling does not seem to align the center of the perimeter of the group with the
        # center of the perimeter of the vertex when of the same exact size. So we are
        # probably ok winging it as users will adapt.
        self._compute_paths(self.get_figure(root=True).dpi)
        super().draw(renderer)


def _compute_group_patch_stub(
    points,
    vertexpadding,
    **kwargs,
):
    if vertexpadding == 0:
        return mpl.patches.Polygon(
            points,
            **kwargs,
        )

    # NOTE: Closing point: mpl is a bit quirky here
    vertices = np.zeros(
        (1 + 30 * len(points), 2),
    )
    codes = ["MOVETO"] + ["LINETO"] * (len(vertices) - 2) + ["CLOSEPOLY"]
    codes = [getattr(mpl.path.Path, x) for x in codes]
    patch = mpl.patches.PathPatch(
        mpl.path.Path(
            vertices,
            codes=codes,
        ),
        **kwargs,
    )

    return patch



from functools import wraps, partial
from math import atan2
import numpy as np
import matplotlib as mpl

from .geometry import (
    _evaluate_squared_bezier,
    _evaluate_cubic_bezier,
)


# NOTE: https://github.com/networkx/grave/blob/main/grave/grave.py
def _stale_wrapper(func):
    """Decorator to manage artist state."""

    @wraps(func)
    def inner(self, *args, **kwargs):
        try:
            func(self, *args, **kwargs)
        finally:
            self.stale = False

    return inner


def _forwarder(forwards, cls=None):
    """Decorator to forward specific methods to Artist children."""
    if cls is None:
        return partial(_forwarder, forwards)

    def make_forward(name):
        def method(self, *args, **kwargs):
            """Each decorated method is called on the decorated class and then, nonrecursively, on all children."""
            ret = getattr(cls.mro()[1], name)(self, *args, **kwargs)
            for c in self.get_children():
                getattr(c, name)(*args, **kwargs)
            return ret

        return method

    for f in forwards:
        method = make_forward(f)
        method.__name__ = f
        method.__doc__ = "broadcasts {} to children".format(f)
        setattr(cls, f, method)

    return cls


def _additional_set_methods(attributes, cls=None):
    """Decorator to add specific set methods for children properties.

    This is useful to autogenerate methods a la set_<key>(value), for
    instance set_alpha(value). It works by delegating to set(alpha=value).

    Overall, this is a minor tweak compared to the previous decorator.
    """
    if cls is None:
        return partial(_additional_set_methods, attributes)

    def make_setter(name):
        def method(self, value):
            self.set(**{name: value})

        return method

    for attr in attributes:
        desc = attr.replace("_", " ")
        method = make_setter(attr)
        method.__name__ = f"set_{attr}"
        method.__doc__ = f"Set {desc}."
        setattr(cls, f"set_{attr}", method)

    return cls


# FIXME: this method appears quite inconsistent, would be better to improve.
# The issue is that to really know the size of a label on screen, we need to
# render it first. Therefore, we should render the labels, then render the
# vertices. Leaving for now, since this can be styled manually which covers
# many use cases.
def _get_label_width_height(text, hpadding=18, vpadding=12, **kwargs):
    """Get the bounding box size for a text with certain properties."""
    forbidden_props = [
        "horizontalalignment",
        "verticalalignment",
        "ha",
        "va",
        "color",
        "edgecolor",
        "facecolor",
    ]
    for prop in forbidden_props:
        if prop in kwargs:
            del kwargs[prop]

    path = mpl.textpath.TextPath((0, 0), text, **kwargs)
    boundingbox = path.get_extents()
    width = boundingbox.width
    height = boundingbox.height

    # Scaling with font size appears broken... try to patch it up linearly here, even though we know it don't work well
    width *= kwargs.get("size", 12) / 12.0
    height *= kwargs.get("size", 12) / 12.0

    width += hpadding
    height += vpadding
    return (width, height)


def _compute_mid_coord_and_rot(path, trans):
    """Compute mid point of an edge, straight or curved."""
    # Distinguish between straight and curved paths
    if path.codes[-1] == mpl.path.Path.LINETO:
        coord = path.vertices.mean(axis=0)
        vtr = trans(path.vertices)
        rot = atan2(
            vtr[-1, 1] - vtr[0, 1],
            vtr[-1, 0] - vtr[0, 0],
        )

    # Cubic Bezier
    elif path.codes[-1] == mpl.path.Path.CURVE4:
        coord = _evaluate_cubic_bezier(path.vertices, 0.5)
        # TODO:
        rot = 0

    # Square Bezier
    elif path.codes[-1] == mpl.path.Path.CURVE3:
        coord = _evaluate_squared_bezier(path.vertices, 0.5)
        # TODO:
        rot = 0

    else:
        raise ValueError(
            "Curve type not straight and not squared/cubic Bezier, cannot compute mid point."
        )

    return coord, rot


def _build_cmap_fun(values, cmap):
    """Map colormap on top of numerical values."""
    cmap = mpl.cm._ensure_cmap(cmap)

    if np.isscalar(values):
        values = [values]

    if isinstance(values, dict):
        values = np.array(list(values.values()))

    vmin = np.nanmin(values)
    vmax = np.nanmax(values)
    norm = mpl.colors.Normalize(vmin=vmin, vmax=vmax)

    return lambda x: cmap(norm(x))



"""
iplotx and other visualisations
===============================

This example shows how to combine `iplotx` visualisations with other tools such as `scipy` and vanilla `matplotlib`. This is one of the key strengths of `iplotx` over custom visualisations in many other packages: `iplotx` produces fully compliant matplotlib artists that can be combined with other matplotlib tools, including `seaborn`.

"""

import igraph as ig
import numpy as np
import matplotlib.pyplot as plt
import iplotx as ipx
from scipy.spatial import Voronoi, voronoi_plot_2d
from scipy.spatial.distance import cdist

# Plot a graph with iplotx
g = ig.Graph.Ring(8, directed=True)
fig, ax = plt.subplots()
art = ipx.plot(
    g,
    ax=ax,
    layout="circle",
    aspect=1,
)[0]

# Add a Voronoi tasselation via scipy
vor = Voronoi(art.get_layout().values)
voronoi_plot_2d(
    vor,
    ax=ax,
    show_points=False,
    show_vertices=False,
    line_colors="tomato",
    line_style="-",
    line_width=2,
)
xmin, xmax = ax.get_xlim()
ymin, ymax = ax.get_ylim()

# Add random points and color them based on Voronoi region
palette = [
    "tomato",
    "steelblue",
    "purple",
    "yellow",
    "deeppink",
    "lawngreen",
    "grey",
    "black",
]
rngpoints = np.random.rand(10000, 2)
rngpoints *= np.array([xmax - xmin, ymax - ymin])
rngpoints += np.array([xmin, ymin])

colors = np.array(palette)[cdist(art.get_layout().values, rngpoints).argmin(axis=0)]
ax.scatter(
    rngpoints[:, 0],
    rngpoints[:, 1],
    c=colors,
    s=20,
    alpha=0.3,
    zorder=0.5,
)



import os
from copy import deepcopy
import unittest
import pytest
import numpy as np

import iplotx as ipx


class StyleTestRunner(unittest.TestCase):
    def test_flat_style(self):
        style = deepcopy(ipx.style.current)

        with ipx.style.context(
            dict(
                vertex_size=80,
                edge_label_bbox_facecolor="yellow",
            ),
        ):
            current = ipx.style.current
            self.assertEqual(current["vertex"]["size"], 80)
            self.assertEqual(current["edge"]["label"]["bbox"]["facecolor"], "yellow")

            with ipx.style.context(
                vertex_size=70,
            ):
                self.assertEqual(current["vertex"]["size"], 70)
                self.assertEqual(
                    current["edge"]["label"]["bbox"]["facecolor"], "yellow"
                )

    def test_generator(self):
        styles = iter([{"vertex_size": 80}, {"vertex_size": 70}])
        with ipx.style.context(
            styles,
        ):
            self.assertEqual(ipx.style.current["vertex"]["size"], 70)

    def test_use(self):
        style = deepcopy(ipx.style.current)
        ipx.style.use("hollow")
        ipx.style.use("default")
        self.assertEqual(style, ipx.style.current)


def suite():
    return unittest.TestSuite(
        [
            unittest.defaultTestLoader.loadTestsFromTestCase(StyleTestRunner),
        ]
    )


def test():
    runner = unittest.TextTestRunner()
    runner.run(suite())


if __name__ == "__main__":
    test()



"""
Maximum Bipartite Matching by Maximum Flow
==========================================

This example from `igraph` presents how to visualise bipartite matching using maximum flow, with edge linewidth and vertex facecolor styling.

"""

import igraph as ig
import matplotlib.pyplot as plt
import iplotx as ipx

# We start by creating the bipartite directed graph.
g = ig.Graph(
    9,
    [(0, 4), (0, 5), (1, 4), (1, 6), (1, 7), (2, 5), (2, 7), (2, 8), (3, 6), (3, 7)],
    directed=True,
)

# We assign:
#  - nodes 0-3 to one side
#  - nodes 4-8 to the other side
g.vs[range(4)]["type"] = True
g.vs[range(4, 9)]["type"] = False

# Then we add a source (vertex 9) and a sink (vertex 10)
g.add_vertices(2)
g.add_edges([(9, 0), (9, 1), (9, 2), (9, 3)])  # connect source to one side
g.add_edges([(4, 10), (5, 10), (6, 10), (7, 10), (8, 10)])  # ... and sinks to the other

# Compute maximum flow
flow = g.maxflow(9, 10)

# To achieve a pleasant visual effect, we set the positions of source and sink
# manually:
layout = g.layout_bipartite()
layout[9] = (2, -1)
layout[10] = (2, 2)

ipx.plot(
    g,
    layout=layout,
    vertex_labels=True,
    style={
        "vertex": {
            "size": 30,
            "facecolor": ["black" if i < 9 else "darkorange" for i in range(11)],
        },
        "edge": {
            "linewidth": [1.0 + flow.flow[i] for i in range(g.ecount())],
        },
    },
)

# %%
# If you want to have dark labels on light background, you can set the
# vertex style accordingly, perhaps with pronounced vertex borders to
# increase constrast:
#
ipx.plot(
    g,
    layout=layout,
    vertex_labels=True,
    style={
        "vertex": {
            "size": 30,
            "facecolor": ["lightblue" if i < 9 else "orange" for i in range(11)],
            "edgecolor": "black",
            "linewidth": 1.5,
            "label": {
                "color": "black",
            },
        },
        "edge": {
            "linewidth": [1.0 + flow.flow[i] for i in range(g.ecount())],
        },
    },
)  #



"""
Heuristics module to funnel certain variable inputs (e.g. layouts) into a standard format.
"""

from typing import (
    Optional,
    Any,
)
from collections.abc import Hashable
from collections import defaultdict
import numpy as np
import pandas as pd

from ..layout import compute_tree_layout
from ..typing import (
    GraphType,
    GroupingType,
    LayoutType,
)


def number_of_vertices(network: GraphType) -> int:
    """Get the number of vertices in the network."""
    from . import network_library

    if network_library(network) == "igraph":
        return network.vcount()
    if network_library(network) == "networkx":
        return network.number_of_nodes()
    raise TypeError("Unsupported graph type. Supported types are igraph and networkx.")


def detect_directedness(
    network: GraphType,
) -> bool:
    """Detect if the network is directed or not."""
    from . import network_library

    nl = network_library(network)

    if nl == "igraph":
        return network.is_directed()
    if nl == "networkx":
        import networkx as nx

        if isinstance(network, (nx.DiGraph, nx.MultiDiGraph)):
            return True
    return False


def normalise_layout(layout, network=None):
    """Normalise the layout to a pandas.DataFrame."""
    from . import network_library

    try:
        import igraph as ig
    except ImportError:
        ig = None

    if layout is None:
        if (network is not None) and (number_of_vertices(network) == 0):
            return pd.DataFrame(np.zeros((0, 2)))
        return None
    if (network is not None) and isinstance(layout, str):
        if network_library(network) == "igraph":
            if hasattr(network, layout):
                layout = network[layout]
            else:
                layout = network.layout(layout)
                # NOTE: This seems like a legit bug in igraph
                # Sometimes (e.g. sugiyama) the layout has more vertices than the network (?)
                layout = np.asarray(layout.coords)[: network.vcount()]
        if network_library(network) == "networkx":
            layout = dict(network.nodes.data(layout))

    if (ig is not None) and isinstance(layout, ig.layout.Layout):
        return pd.DataFrame(layout.coords)
    if isinstance(layout, dict):
        return pd.DataFrame(layout).T
    if isinstance(layout, str):
        raise NotImplementedError("Layout as a string is not supported yet.")
    if isinstance(layout, (list, tuple)):
        return pd.DataFrame(np.array(layout))
    if isinstance(layout, pd.DataFrame):
        return layout
    if isinstance(layout, np.ndarray):
        return pd.DataFrame(layout)
    raise TypeError("Layout could not be normalised.")


def normalise_tree_layout(
    layout: str | Any,
    **kwargs,
) -> pd.DataFrame:
    """Normalise tree layout from a variety of inputs.

    Parameters:
        layout: The tree layout to normalise.
        tree: The correcponding tree object.
        **kwargs: Additional arguments for the subroutines.

    Returns:
        A pandas DataFrame with the normalised tree layout.

    NOTE: This function currently only accepts strings and computes
        the layout internally. This might change in the future.
    """
    if isinstance(layout, str):
        layout = compute_tree_layout(layout, **kwargs)
    else:
        raise NotImplementedError(
            "Only internally computed tree layout currently accepted."
        )

    if isinstance(layout, dict):
        # Adjust vertex layout
        index = []
        coordinates = []
        for key, coordinate in layout.items():
            index.append(key)
            coordinates.append(coordinate)
        index = pd.Index(index)
        coordinates = np.array(coordinates)
        ndim = len(coordinates[0]) if len(coordinates) > 0 else 2
        layout_columns = [f"_ipx_layout_{i}" for i in range(ndim)]
        layout = pd.DataFrame(
            coordinates,
            index=index,
            columns=layout_columns,
        )

    return layout


def normalise_grouping(
    grouping: GroupingType,
    layout: LayoutType,
) -> dict[Hashable, set]:
    """Normalise network grouping from a variery of inputs.

    Parameters:
        grouping: Network grouping (e.g. vertex cover).
        layout: Network layout.

    Returns:
        A dictionary of sets. Each key is the index of a group, each value is a set of vertices
        included in that group. If all sets are mutually exclusive, this is a vertex clustering,
        otherwise it's only a vertex cover.
    """
    try:
        import igraph as ig
    except ImportError:
        ig = None

    if len(grouping) == 0:
        return {}

    if isinstance(grouping, dict):
        val0 = next(iter(grouping.values()))
        # If already the right data type or compatible, leave as is
        if isinstance(val0, (set, frozenset)):
            return grouping

        # If a dict of integers or strings, assume each key is a vertex id and each value is a
        # group, convert (i.e. invert the dict)
        if isinstance(val0, (int, str)):
            group_dic = defaultdict(set)
            for key, val in grouping.items():
                group_dic[val].add(key)
            return group_dic

    # If an igraph object, convert to a dict of sets
    if ig is not None:
        if isinstance(grouping, ig.clustering.Clustering):
            layout = normalise_layout(layout)
            group_dic = defaultdict(set)
            for i, member in enumerate(grouping.membership):
                group_dic[member].add(i)
            return group_dic

        if isinstance(grouping, ig.clustering.Cover):
            layout = normalise_layout(layout)
            group_dic = defaultdict(set)
            for i, members in enumerate(grouping.membership):
                for member in members:
                    group_dic[member].add(i)
            return group_dic

    # Assume it's a sequence, so convert to list
    grouping = list(grouping)

    # If the values are already sets, assume group indices are integers
    # and values are as is
    if isinstance(grouping[0], set):
        return dict(enumerate(grouping))

    # If the values are integers or strings, assume each key is a vertex id and each value is a
    # group, convert to dict of sets
    if isinstance(grouping[0], (int, str)):
        group_dic = defaultdict(set)
        for i, val in enumerate(grouping):
            group_dic[val].add(i)
        return group_dic

    raise TypeError(
        "Could not standardise grouping from object.",
    )



"""
Module for label collection in iplotx.
"""

from typing import (
    Optional,
    Sequence,
)
import numpy as np
import pandas as pd
import matplotlib as mpl

from .style import (
    rotate_style,
    copy_with_deep_values,
)
from .utils.matplotlib import (
    _stale_wrapper,
    _forwarder,
)


@_forwarder(
    (
        "set_clip_path",
        "set_clip_box",
        "set_snap",
        "set_sketch_params",
        "set_animated",
        "set_picker",
    )
)
class LabelCollection(mpl.artist.Artist):
    """Collection of labels for iplotx with styles.

    NOTE: This class is not a subclass of `mpl.collections.Collection`, although in some ways items
    behaves like one. It is named LabelCollection quite literally to indicate it contains a list of
    labels for vertices, edges, etc.
    """

    def __init__(
        self,
        labels: pd.Series,
        style: Optional[dict[str, dict]] = None,
        offsets: Optional[np.ndarray] = None,
        transform: mpl.transforms.Transform = mpl.transforms.IdentityTransform(),
    ) -> None:
        """Initialize a collection of labels.

        Parameters:
            labels: A sequence of labels to be displayed.
            style: A dictionary of styles to apply to the labels. The keys are style properties.
            offsets: A sequence of offsets for each label, specifying the position of the label.
            transform: A transform to apply to the labels. This is usually ax.transData.
        """
        self._labels = labels
        self._offsets = offsets if offsets is not None else np.zeros((len(labels), 2))
        self._style = style
        super().__init__()

        self.set_transform(transform)
        self._create_artists()

    def get_children(self) -> tuple[mpl.artist.Artist]:
        """Get the children of this artist, which are the label artists."""
        return tuple(self._labelartists)

    def set_figure(self, fig) -> None:
        """Set the figure of this artist.

        Parameters:
            fig: The figure to set.
        """
        super().set_figure(fig)
        for child in self.get_children():
            child.set_figure(fig)
        self._update_offsets(dpi=fig.dpi)

    def _get_margins_with_dpi(self, dpi: float = 72.0) -> np.ndarray:
        return self._margins * dpi / 72.0

    def _create_artists(self) -> None:
        style = copy_with_deep_values(self._style) if self._style is not None else {}
        transform = self.get_transform()

        margins = []

        forbidden_props = ["rotate"]
        for prop in forbidden_props:
            if prop in style:
                del style[prop]

        arts = []
        for i, (anchor_id, label) in enumerate(self._labels.items()):
            stylei = rotate_style(style, index=i, key=anchor_id)
            # Margins are handled separately
            hmargin = stylei.pop("hmargin", 0.0)
            vmargin = stylei.pop("vmargin", 0.0)
            margins.append((hmargin, vmargin))

            # Initially, ignore autoalignment since we do not know the
            # rotations
            if stylei.get("horizontalalignment") == "auto":
                stylei["horizontalalignment"] = "center"

            art = mpl.text.Text(
                self._offsets[i][0],
                self._offsets[i][1],
                label,
                transform=transform,
                **stylei,
            )
            arts.append(art)
        self._labelartists = arts
        self._margins = np.array(margins)
        self._rotations = np.zeros(len(self._labels))

    def _update_offsets(self, dpi: float = 72.0) -> None:
        """Update offsets including margins."""
        self.set_offsets(self._offsets, dpi=dpi)

    def get_offsets(self, with_margins: bool = False) -> np.ndarray:
        """Get the positions (offsets) of the labels."""
        if not with_margins:
            return self._offsets
        else:
            return np.array(
                [art.get_position() for art in self._labelartists],
            )

    def _adjust_offsets_for_margins(self, offsets, dpi=72.0):
        margins = self._get_margins_with_dpi(dpi=dpi)
        if (margins != 0).any():
            transform = self.get_transform()
            trans = transform.transform
            trans_inv = transform.inverted().transform
            rotations = self.get_rotations()
            vrot = [np.cos(rotations), np.sin(rotations)]

            margins_rot = np.empty_like(margins)
            margins_rot[:, 0] = margins[:, 0] * vrot[0] - margins[:, 1] * vrot[1]
            margins_rot[:, 1] = margins[:, 0] * vrot[1] + margins[:, 1] * vrot[0]
            offsets = trans_inv(trans(offsets) + margins_rot)
        return offsets

    def set_offsets(self, offsets, dpi: float = 72.0) -> None:
        """Set positions (offsets) of the labels.

        Parameters:
            offsets: A sequence of offsets for each label, specifying the position of the label.
        """
        self._offsets = np.asarray(offsets)
        offsets_with_margins = self._adjust_offsets_for_margins(offsets, dpi=dpi)
        for art, offset in zip(self._labelartists, offsets_with_margins):
            art.set_position((offset[0], offset[1]))

    def get_rotations(self) -> np.ndarray:
        """Get the rotations of the labels in radians."""
        return self._rotations

    def set_rotations(self, rotations: Sequence[float]) -> None:
        """Set the rotations of the labels.

        Parameters:
            rotations: A sequence of rotations in radians for each label.
        """
        self._rotations = np.asarray(rotations)
        ha = self._style.get("horizontalalignment", "center")
        for art, rotation in zip(self._labelartists, rotations):
            rot_deg = 180.0 / np.pi * rotation
            # Force the font size to be upwards
            if ha == "auto":
                if -90 <= rot_deg < 90:
                    art.set_horizontalalignment("left")
                else:
                    art.set_horizontalalignment("right")
            rot_deg = ((rot_deg + 90) % 180) - 90
            art.set_rotation(rot_deg)

    def get_datalim(self, transData=None) -> mpl.transforms.Bbox:
        """Get the data limits of the labels."""
        bboxes = self.get_datalims_children(transData=transData)
        bbox = mpl.transforms.Bbox.union(bboxes)
        return bbox

    def get_datalims_children(self, transData=None) -> Sequence[mpl.transforms.Bbox]:
        """Get the data limits of the children of this artist."""
        if transData is None:
            transData = self.get_transform()
        trans_inv = transData.inverted().transform_bbox
        bboxes = []
        for art in self._labelartists:
            bbox_fig = art.get_bbox_patch().get_extents()
            bbox_data = trans_inv(bbox_fig)
            bboxes.append(bbox_data)
        return bboxes

    @_stale_wrapper
    def draw(self, renderer) -> None:
        """Draw each of the children, with some buffering mechanism."""
        if not self.get_visible():
            return

        self._update_offsets(dpi=renderer.dpi)

        # We should manage zorder ourselves, but we need to compute
        # the new offsets and angles of arrows from the edges before drawing them
        for art in self.get_children():
            art.draw(renderer)



"""
Basic network plotting
======================

This example shows how to plot a simple graph with `iplotx`.
"""

import igraph as ig
import matplotlib.pyplot as plt
import iplotx as ipx

g = ig.Graph.Ring(5)
layout = g.layout("circle").coords
fig, ax = plt.subplots(figsize=(3, 3))
ipx.plot(g, ax=ax, layout=layout)



"""
Colorbar
========

This example from networkx shows how to use mappable colors and colormaps and how to connect them to a colorbar.
"""

import networkx as nx
import matplotlib.pyplot as plt
import iplotx as ipx

seed = 13648  # Seed random number generators for reproducibility
G = nx.random_k_out_graph(10, 3, 0.5, seed=seed)
pos = nx.spring_layout(G, seed=seed)

node_sizes = [3 + 1.5 * i for i in range(len(G))]
M = G.number_of_edges()
edge_colors = list(range(2, M + 2))
edge_alphas = [(5 + i) / (M + 4) for i in range(M)]
cmap = plt.cm.plasma

fig, ax = plt.subplots(figsize=(6, 4))

network_artist = ipx.plot(
    network=G,
    ax=ax,
    layout=pos,
    style={
        "vertex": {
            "size": node_sizes,
            "facecolor": "indigo",
            "edgecolor": "none",
        },
        "edge": {
            "color": edge_colors,
            "alpha": edge_alphas,
            "cmap": cmap,
            "linewidth": 2,
            "offset": 0,
            "arrow": {
                "marker": ">",
                "width": 5,
            },
        },
    },
)[0]
fig.colorbar(
    network_artist.get_edges(),
    ax=ax,
)

# %%
# The arrow colors can be specified explicitely, in which case it overrides the edge color:

fig, ax = plt.subplots(figsize=(6, 4))

network_artist = ipx.plot(
    network=G,
    ax=ax,
    layout=pos,
    style={
        "vertex": {
            "size": node_sizes,
            "facecolor": "indigo",
            "edgecolor": "none",
        },
        "edge": {
            "color": edge_colors,
            "alpha": edge_alphas,
            "cmap": cmap,
            "linewidth": 2,
            "offset": 0,
            "arrow": {
                "marker": ">",
                "width": 5,
                "color": "black",  # Make arrows black
            },
        },
    },
)[0]
fig.colorbar(
    network_artist.get_edges(),
    ax=ax,
)

# %%
# The transparency (alpha) of the arrows is still determined by the edges unless specified:

fig, ax = plt.subplots(figsize=(6, 4))

network_artist = ipx.plot(
    network=G,
    ax=ax,
    layout=pos,
    style={
        "vertex": {
            "size": node_sizes,
            "facecolor": "indigo",
            "edgecolor": "none",
        },
        "edge": {
            "color": edge_colors,
            "alpha": edge_alphas,
            "cmap": cmap,
            "linewidth": 2,
            "offset": 0,
            "arrow": {
                "marker": ">",
                "width": 5,
                "color": "black",  # Make arrows black
                "alpha": 1.0,  # Set 100% opacity for arrows
            },
        },
    },
)[0]
fig.colorbar(
    network_artist.get_edges(),
    ax=ax,
)



from typing import (
    Any,
    Optional,
    Sequence,
)
from ...typing import (
    TreeDataProvider,
)


class Cogent3DataProvider(TreeDataProvider):
    def preorder(self) -> Sequence[Any]:
        return self.tree.preorder()

    def postorder(self) -> Sequence[Any]:
        return self.tree.postorder()

    def get_leaves(self) -> Sequence[Any]:
        return self.tree.tips()

    @staticmethod
    def get_children(node: Any) -> Sequence[Any]:
        return node.children

    @staticmethod
    def get_branch_length(node: Any) -> Optional[float]:
        return node.length

    @staticmethod
    def check_dependencies() -> bool:
        try:
            import cogent3
        except ImportError:
            return False
        return True

    @staticmethod
    def tree_type():
        from cogent3.core.tree import PhyloNode

        return PhyloNode



"""
Mouse event handling
====================

This example shows how to interact with mouse events (e.g. hovering) in `iplotx`.

.. warning::
  This example will run in a Python, IPython, or Jupyter session, however the
  interactive functionality is not visible on the HTML page. Download the code
  at the end of this page and run it in a local Python environment to see
  the results.
"""

import matplotlib.pyplot as plt
import numpy as np
import igraph as ig
import iplotx as ipx

g = ig.Graph.Ring(3, directed=True)

fig, ax = plt.subplots()
art = ipx.network(
    g,
    layout="circle",
    ax=ax,
    aspect=1,
)[0]
vertex_artist = art.get_vertices()

# Prepare an invisible annotation for hovering
annot = ax.annotate(
    "",
    xy=(0, 0),
    xytext=(20, 20),
    textcoords="offset points",
    bbox=dict(boxstyle="round", fc=(0, 0, 0, 0.2)),
    arrowprops=dict(arrowstyle="->"),
)
annot.set_visible(False)


def hover_callback(event):
    """React to mouse hovering over vertices."""
    if event.inaxes == ax:
        vc = art.get_vertices()
        cont, ind = vc.contains(event)
        # If mouse is over a vertex, show the buble
        if cont:
            i = ind["ind"][0]
            annot.xy = vc.get_offsets()[i]
            annot.set_text(f"{i+1}")
            annot.set_visible(True)
        # Otherwise, hide the bubble
        elif annot.get_visible():
            annot.set_visible(False)
        # If nothing changed, no need to redraw
        else:
            return

        # Redraw to show/hide the bubble
        fig.canvas.draw_idle()


fig.canvas.mpl_connect(
    "motion_notify_event",
    hover_callback,
)



from typing import (
    Optional,
    Sequence,
)
from collections.abc import Hashable
from collections import defaultdict

import numpy as np
import pandas as pd
import matplotlib as mpl

from .style import (
    context,
    get_style,
    rotate_style,
)
from .utils.matplotlib import (
    _stale_wrapper,
    _forwarder,
    _build_cmap_fun,
)
from .ingest import (
    ingest_tree_data,
    data_providers,
)
from .vertex import (
    VertexCollection,
)
from .edge import (
    EdgeCollection,
    make_stub_patch as make_undirected_edge_patch,
)
from .label import (
    LabelCollection,
)
from .cascades import (
    CascadeCollection,
)
from .network import (
    _update_from_internal,
)


@_forwarder(
    (
        "set_clip_path",
        "set_clip_box",
        "set_snap",
        "set_sketch_params",
        "set_animated",
        "set_picker",
    )
)
class TreeArtist(mpl.artist.Artist):
    """Artist for plotting trees."""

    def __init__(
        self,
        tree,
        layout: Optional[str] = "horizontal",
        orientation: Optional[str] = None,
        directed: bool | str = False,
        vertex_labels: Optional[
            bool | list[str] | dict[Hashable, str] | pd.Series
        ] = None,
        edge_labels: Optional[Sequence | dict[Hashable, str] | pd.Series] = None,
        leaf_labels: Optional[Sequence | dict[Hashable, str]] | pd.Series = None,
        transform: mpl.transforms.Transform = mpl.transforms.IdentityTransform(),
        offset_transform: Optional[mpl.transforms.Transform] = None,
    ):
        """Initialize the TreeArtist.

        Parameters:
            tree: The tree to plot.
            layout: The layout to use for the tree. Can be "horizontal", "vertical", or "radial".
            orientation: The orientation of the tree layout. Can be "right" or "left" (for
                horizontal and radial layouts) and "descending" or "ascending" (for vertical
                layouts).
            directed: Whether the tree is directed. Can be a boolean or a string with the
                following choices: "parent" or "child".
            vertex_labels: Labels for the vertices. Can be a list, dictionary, or pandas Series.
            edge_labels: Labels for the edges. Can be a sequence of strings.
            leaf_labels: Labels for the leaves. Can be a sequence of strings or a pandas Series.
                These labels are positioned at the depth of the deepest leaf. If you want to
                label leaves next to each leaf independently of how deep they are, use
                the "vertex_labels" parameter instead - usually as a dict with the leaves
                as keys and the labels as values.
            transform: The transform to apply to the tree artist. This is usually the identity.
            offset_transform: The offset transform to apply to the tree artist. This is
                usually `ax.transData`.
        """

        self.tree = tree
        self._ipx_internal_data = ingest_tree_data(
            tree,
            layout,
            orientation=orientation,
            directed=directed,
            layout_style=get_style(".layout", {}),
            vertex_labels=vertex_labels,
            edge_labels=edge_labels,
            leaf_labels=leaf_labels,
        )

        super().__init__()

        # This is usually the identity (which scales poorly with dpi)
        self.set_transform(transform)

        # This is usually transData
        self.set_offset_transform(offset_transform)

        zorder = get_style(".network").get("zorder", 1)
        self.set_zorder(zorder)

        self._add_vertices()
        self._add_edges()
        self._add_leaf_vertices()

        # NOTE: cascades need to be created after leaf vertices in case
        # they are requested to wrap around them.
        if "cascade" in self.get_vertices().get_style():
            self._add_cascades()

    def get_children(self) -> tuple[mpl.artist.Artist]:
        """Get the children of this artist.

        Returns:
            The artists for vertices and edges.
        """
        children = [self._vertices, self._edges]
        if hasattr(self, "_leaf_vertices"):
            children.append(self._leaf_vertices)
        if hasattr(self, "_cascades"):
            children.append(self._cascades)
        return tuple(children)

    def set_figure(self, fig) -> None:
        """Set the figure for this artist and its children.

        Parameters:
            fig: the figure to set for this artist and its children.
        """
        super().set_figure(fig)
        for child in self.get_children():
            child.set_figure(fig)

        # At the end, if there are cadcades with extent depending on
        # leaf edges, we should update them
        self._update_cascades_extent()

    def _update_cascades_extent(self) -> None:
        """Update cascades if extent depends on leaf labels."""
        if not hasattr(self, "_cascades"):
            return

        style_cascade = self.get_vertices().get_style()["cascade"]
        extend_to_labels = style_cascade.get("extend", False) == "leaf_labels"
        if not extend_to_labels:
            return

        maxdepth = self._get_maxdepth_leaf_labels()
        self._cascades.set_maxdepth(maxdepth)

    def get_offset_transform(self):
        """Get the offset transform (for vertices/edges)."""
        return self._offset_transform

    def set_offset_transform(self, offset_transform):
        """Set the offset transform (for vertices/edges)."""
        self._offset_transform = offset_transform

    def get_layout(self, kind="vertex"):
        """Get vertex or edge layout."""
        layout_columns = [
            f"_ipx_layout_{i}" for i in range(self._ipx_internal_data["ndim"])
        ]

        if kind == "vertex":
            layout = self._ipx_internal_data["vertex_df"][layout_columns]
            return layout
        elif kind == "leaf":
            leaves = self._ipx_internal_data["leaf_df"].index
            layout = self._ipx_internal_data["vertex_df"][layout_columns]
            # NOTE: workaround for a pandas bug
            idxs = []
            for i, vid in enumerate(layout.index):
                if vid in leaves:
                    idxs.append(i)
            layout = layout.iloc[idxs]
            return layout

        elif kind == "edge":
            return self._ipx_internal_data["edge_df"][layout_columns]
        else:
            raise ValueError(f"Unknown layout kind: {kind}. Use 'vertex' or 'edge'.")

    def get_datalim(self, transData, pad=0.15):
        """Get limits on x/y axes based on the graph layout data.

        Parameters:
            transData (Transform): The transform to use for the data.
            pad (float): Padding to add to the limits. Default is 0.05.
                Units are a fraction of total axis range before padding.
        """
        layout = self.get_layout().values

        if len(layout) == 0:
            return mpl.transforms.Bbox([[0, 0], [1, 1]])

        bbox = self._vertices.get_datalim(transData)

        edge_bbox = self._edges.get_datalim(transData)
        bbox = mpl.transforms.Bbox.union([bbox, edge_bbox])

        if hasattr(self, "_cascades"):
            cascades_bbox = self._cascades.get_datalim(transData)
            bbox = mpl.transforms.Bbox.union([bbox, cascades_bbox])

        if hasattr(self, "_leaf_vertices"):
            leaf_labels_bbox = self._leaf_vertices.get_datalim(transData)
            bbox = mpl.transforms.Bbox.union([bbox, leaf_labels_bbox])

        bbox = bbox.expanded(sw=(1.0 + pad), sh=(1.0 + pad))
        return bbox

    def _get_label_series(self, kind: str) -> Optional[pd.Series]:
        if "label" in self._ipx_internal_data[f"{kind}_df"].columns:
            return self._ipx_internal_data[f"{kind}_df"]["label"]
        else:
            return None

    def get_vertices(self) -> VertexCollection:
        """Get VertexCollection artist."""
        return self._vertices

    def get_edges(self) -> EdgeCollection:
        """Get EdgeCollection artist."""
        return self._edges

    def get_leaf_vertices(self) -> Optional[VertexCollection]:
        """Get leaf VertexCollection artist."""
        if hasattr(self, "_leaf_vertices"):
            return self._leaf_vertices
        return None

    def get_vertex_labels(self) -> LabelCollection:
        """Get list of vertex label artists."""
        return self._vertices.get_labels()

    def get_edge_labels(self) -> LabelCollection:
        """Get list of edge label artists."""
        return self._edges.get_labels()

    def get_leaf_labels(self) -> Optional[LabelCollection]:
        if hasattr(self, "_leaf_vertices"):
            return self._leaf_vertices.get_labels()
        return None

    def _add_vertices(self) -> None:
        """Add vertices to the tree."""
        self._vertices = VertexCollection(
            layout=self.get_layout(),
            layout_coordinate_system=self._ipx_internal_data.get(
                "layout_coordinate_system",
                "catesian",
            ),
            style=get_style(".vertex"),
            labels=self._get_label_series("vertex"),
            transform=self.get_transform(),
            offset_transform=self.get_offset_transform(),
        )

    def _add_leaf_vertices(self) -> None:
        """Add invisible deep vertices as leaf label anchors."""
        leaf_layout = self.get_layout("leaf").copy()
        # Set all to max depth
        depth_idx = int(self._ipx_internal_data["layout_name"] == "vertical")
        leaf_layout.iloc[:, depth_idx] = leaf_layout.iloc[:, depth_idx].max()

        # Set invisible vertices with visible labels
        layout_name = self._ipx_internal_data["layout_name"]
        orientation = self._ipx_internal_data["orientation"]
        if layout_name == "radial":
            ha = "auto"
        elif orientation in ("left", "ascending"):
            ha = "right"
        else:
            ha = "left"

        leaf_vertex_style = {
            "size": 0,
            "label": {
                "verticalalignment": "center",
                "horizontalalignment": ha,
                "hmargin": 5,
                "bbox": {
                    "facecolor": (1, 1, 1, 0),
                },
            },
        }
        with context({"vertex": leaf_vertex_style}):
            leaf_vertex_style = get_style(".vertex")
            self._leaf_vertices = VertexCollection(
                layout=leaf_layout,
                layout_coordinate_system=self._ipx_internal_data.get(
                    "layout_coordinate_system",
                    "catesian",
                ),
                style=leaf_vertex_style,
                labels=self._get_label_series("leaf"),
                transform=self.get_transform(),
                offset_transform=self.get_offset_transform(),
            )

    def _add_cascades(self) -> None:
        """Add cascade patches."""
        # NOTE: If leaf labels are present and the cascades are requested to wrap around them,
        # we have to compute the max extend of the cascades from the leaf labels.
        maxdepth = None
        style_cascade = self.get_vertices().get_style()["cascade"]
        extend_to_labels = style_cascade.get("extend", False) == "leaf_labels"
        has_leaf_labels = self.get_leaf_labels() is not None
        if extend_to_labels and not has_leaf_labels:
            raise ValueError("Cannot extend cascades: no leaf labels.")

        if extend_to_labels and has_leaf_labels:
            maxdepth = self._get_maxdepth_leaf_labels()

        self._cascades = CascadeCollection(
            tree=self.tree,
            layout=self.get_layout(),
            layout_name=self._ipx_internal_data["layout_name"],
            orientation=self._ipx_internal_data["orientation"],
            style=style_cascade,
            provider=data_providers["tree"][self._ipx_internal_data["tree_library"]],
            transform=self.get_offset_transform(),
            maxdepth=maxdepth,
        )

    def _get_maxdepth_leaf_labels(self):
        layout_name = self.get_layout_name()
        if layout_name == "radial":
            maxdepth = 0
            # These are the text boxes, they must all be included
            bboxes = self.get_leaf_labels().get_datalims_children(
                self.get_offset_transform()
            )
            for bbox in bboxes:
                r1 = np.linalg.norm([bbox.xmax, bbox.ymax])
                r2 = np.linalg.norm([bbox.xmax, bbox.ymin])
                r3 = np.linalg.norm([bbox.xmin, bbox.ymax])
                r4 = np.linalg.norm([bbox.xmin, bbox.ymin])
                maxdepth = max(maxdepth, r1, r2, r3, r4)
        else:
            orientation = self.get_orientation()
            bbox = self.get_leaf_labels().get_datalim(self.get_offset_transform())
            if (layout_name, orientation) == ("horizontal", "right"):
                maxdepth = bbox.xmax
            elif layout_name == "horizontal":
                maxdepth = bbox.xmin
            elif (layout_name, orientation) == ("vertical", "descending"):
                maxdepth = bbox.ymin
            elif layout_name == "vertical":
                maxdepth = bbox.ymax

        return maxdepth

    def _add_edges(self) -> None:
        """Add edges to the network artist.

        NOTE: UndirectedEdgeCollection and ArrowCollection are both subclasses of
        PatchCollection. When used with a cmap/norm, they set their facecolor
        according to the cmap, even though most likely we only want the edgecolor
        set that way. It can make for funny looking plots that are not uninteresting
        but mostly niche at this stage. Therefore we sidestep the whole cmap thing
        here.
        """

        labels = self._get_label_series("edge")
        edge_style = get_style(".edge")

        if "cmap" in edge_style:
            cmap_fun = _build_cmap_fun(
                edge_style["color"],
                edge_style["cmap"],
            )
        else:
            cmap_fun = None

        edge_df = self._ipx_internal_data["edge_df"].set_index(
            ["_ipx_source", "_ipx_target"]
        )

        if "cmap" in edge_style:
            colorarray = []
        edgepatches = []
        adjacent_vertex_ids = []
        waypoints = []
        for i, (vid1, vid2) in enumerate(edge_df.index):
            edge_stylei = rotate_style(edge_style, index=i, key=(vid1, vid2))

            # FIXME:: Improve this logic. We have three layers of priority:
            # 1. Explicitely set in the style of "plot"
            # 2. Internal through network attributes
            # 3. Default styles
            # Because 1 and 3 are merged as a style context on the way in,
            # it's hard to squeeze 2 in the middle. For now, we will assume
            # the priority order is 2-1-3 instead (internal property is
            # highest priority).
            # This is also why we cannot shift this logic further into the
            # EdgeCollection class, which is oblivious of NetworkArtist's
            # internal data. In fact, one would argue this needs to be
            # pushed outwards to deal with the wrong ordering.
            _update_from_internal(edge_stylei, edge_df.iloc[i], kind="edge")

            if cmap_fun is not None:
                colorarray.append(edge_stylei["color"])
                edge_stylei["color"] = cmap_fun(edge_stylei["color"])

            # Tree layout determines waypoints
            waypointsi = edge_stylei.pop("waypoints", None)
            if waypointsi is None:
                layout_name = self._ipx_internal_data["layout_name"]
                if layout_name == "horizontal":
                    waypointsi = "x0y1"
                elif layout_name == "vertical":
                    waypointsi = "y0x1"
                elif layout_name == "radial":
                    waypointsi = "r0a1"
                else:
                    waypointsi = "none"
            waypoints.append(waypointsi)

            # These are not the actual edges drawn, only stubs to establish
            # the styles which are then fed into the dynamic, optimised
            # factory (the collection) below
            patch = make_undirected_edge_patch(
                **edge_stylei,
            )
            edgepatches.append(patch)
            adjacent_vertex_ids.append((vid1, vid2))

        if "cmap" in edge_style:
            vmin = np.min(colorarray)
            vmax = np.max(colorarray)
            norm = mpl.colors.Normalize(vmin=vmin, vmax=vmax)
            edge_style["norm"] = norm

        edge_style["waypoints"] = waypoints

        # NOTE: Trees are directed is their "directed" property is True, "child", or "parent"
        self._edges = EdgeCollection(
            edgepatches,
            vertex_ids=adjacent_vertex_ids,
            vertex_collection=self._vertices,
            labels=labels,
            transform=self.get_offset_transform(),
            style=edge_style,
            directed=bool(self._ipx_internal_data["directed"]),
        )
        if "cmap" in edge_style:
            self._edges.set_array(colorarray)

    def get_layout_name(self) -> str:
        """Get the layout name."""
        return self._ipx_internal_data["layout_name"]

    def get_orientation(self) -> Optional[str]:
        """Get the orientation of the tree layout."""
        return self._ipx_internal_data.get("orientation", None)

    @_stale_wrapper
    def draw(self, renderer) -> None:
        """Draw each of the children, with some buffering mechanism."""
        if not self.get_visible():
            return

        # At the end, if there are cadcades with extent depending on
        # leaf edges, we should update them
        self._update_cascades_extent()

        # NOTE: looks like we have to manage the zorder ourselves
        # this is kind of funny actually. Btw we need to ensure
        # that cascades are drawn behind (earlier than) vertices
        # and edges at equal zorder because it looks better that way.
        z_suborder = defaultdict(int)
        if hasattr(self, "_cascades"):
            z_suborder[self._cascades] = -1
        children = list(self.get_children())
        children.sort(key=lambda x: (x.zorder, z_suborder[x]))
        for art in children:
            art.draw(renderer)



from typing import (
    Any,
    Optional,
    Sequence,
)
from collections.abc import Hashable
import copy
from contextlib import contextmanager
import numpy as np
import pandas as pd


style_leaves = (
    "cmap",
    "color",
    "size",
    "edgecolor",
    "facecolor",
    "linewidth",
    "linestyle",
    "alpha",
    "zorder",
    "tension",
    "looptension",
    "loopmaxangle",
    "paralleloffset",
    "offset",
    "rotate",
    "marker",
    "waypoints",
    "horizontalalignment",
    "verticalalignment",
    "boxstyle",
    "hpadding",
    "vpadding",
    "hmargin",
    "vmargin",
    "ports",
    "extend",
)

# These properties are not allowed to be rotated (global throughout the graph).
# This might change in the future as the API improves.
nonrotating_leaves = (
    "paralleloffset",
    "looptension",
    "loopmaxangle",
    "vertexpadding",
    "extend",
)


default = {
    "vertex": {
        "size": 20,
        "facecolor": "black",
        "marker": "o",
        "label": {
            "color": "white",
            "horizontalalignment": "center",
            "verticalalignment": "center",
            "hpadding": 18,
            "vpadding": 12,
        },
    },
    "edge": {
        "linewidth": 1.5,
        "linestyle": "-",
        "color": "black",
        "curved": False,
        "paralleloffset": 3,
        "tension": 1,
        "looptension": 4,
        "loopmaxangle": 60,
        "label": {
            "horizontalalignment": "center",
            "verticalalignment": "center",
            "rotate": False,
            "bbox": {
                "boxstyle": "round",
                "facecolor": "white",
                "edgecolor": "none",
            },
        },
        "arrow": {
            "marker": "|>",
            "width": 8,
        },
    },
    "grouping": {
        "facecolor": ["grey", "steelblue", "tomato"],
        "edgecolor": "black",
        "linewidth": 1.5,
        "alpha": 0.5,
        "vertexpadding": 18,
    },
}


def copy_with_deep_values(style):
    """Make a deep copy of the style dict but do not create copies of the keys."""
    newdict = {}
    for key, value in style.items():
        if isinstance(value, dict):
            newdict[key] = copy_with_deep_values(value)
        else:
            newdict[key] = copy.copy(value)
    return newdict


hollow = copy_with_deep_values(default)
hollow["vertex"]["color"] = None
hollow["vertex"]["facecolor"] = "none"
hollow["vertex"]["edgecolor"] = "black"
hollow["vertex"]["linewidth"] = 1.5
hollow["vertex"]["marker"] = "r"
hollow["vertex"]["size"] = "label"
hollow["vertex"]["label"]["color"] = "black"

tree = copy_with_deep_values(default)
tree["vertex"]["size"] = 0
tree["vertex"]["alpha"] = 0
tree["edge"]["linewidth"] = 2.5
tree["vertex"]["label"]["bbox"] = {
    "boxstyle": "square,pad=0.5",
    "facecolor": "white",
    "edgecolor": "none",
}
tree["vertex"]["label"]["color"] = "black"
tree["vertex"]["label"]["size"] = 12
tree["vertex"]["label"]["verticalalignment"] = "center"
tree["vertex"]["label"]["hmargin"] = 10


styles = {
    "default": default,
    "hollow": hollow,
    "tree": tree,
}


stylename = "default"


current = copy_with_deep_values(styles["default"])


def get_stylename():
    """Return the name of the current iplotx style."""
    return str(stylename)


def get_style(name: str = "", *args) -> dict[str, Any]:
    """Get a *deep copy* of the chosen style.

    Parameters:
        name: The name of the style to get. If empty, the current style is returned.
            Substyles can be obtained by using a dot notation, e.g. "default.vertex".
            If "name" starts with a dot, it means a substyle of the current style.
        *args: A single argument is accepted. If present, this value (usually a
            dictionary) is returned if the queried style is not found. For example,
            get_style(".nonexistent") raises an Exception but
            get_style("nonexistent", {}) does not, returning an empty dict instead.
    Returns:
        The requected style or substyle.

    NOTE: The deep copy is a little different from standard deep copies. Here, keys
        (which need to be hashables) are never copied, but values are. This can be
        useful for hashables that change hash upon copying, such as Biopython's
        tree nodes.
    """
    if len(args) > 1:
        raise ValueError("get_style() accepts at most one additional argument.")

    namelist = name.split(".")
    style = styles
    for i, namei in enumerate(namelist):
        if (i == 0) and (namei == ""):
            style = current
        else:
            if namei in style:
                style = style[namei]
            # NOTE: if asking for a nonexistent, non-leaf style
            # give the benefit of the doubt and set an empty dict
            # which will not fail unless the uder tries to enter it
            elif namei not in style_leaves:
                style = {}
            elif len(args) > 0:
                return args[0]
            else:
                raise KeyError(f"Style not found: {name}")

    style = copy_with_deep_values(style)
    return style


# The following is inspired by matplotlib's style library
# https://github.com/matplotlib/matplotlib/blob/v3.10.3/lib/matplotlib/style/core.py#L45
def use(style: Optional[str | dict | Sequence] = None, **kwargs):
    """Use iplotx style setting for a style specification.

    The style name of 'default' is reserved for reverting back to
    the default style settings.

    Parameters:
        style: A style specification, currently either a name of an existing style
            or a dict with specific parts of the style to override. The string
            "default" resets the style to the default one. If this is a sequence,
            each style is applied in order.
        **kwargs: Additional style changes to be applied at the end of any style.
    """
    try:
        import networkx as nx
    except ImportError:
        nx = None

    global current

    def _sanitize_leaves(style: dict):
        for key, value in style.items():
            if key in style_leaves:
                if nx is not None:
                    if isinstance(value, nx.classes.reportviews.NodeView):
                        style[key] = dict(value)
                    elif isinstance(value, nx.classes.reportviews.EdgeViewABC):
                        style[key] = [v for *e, v in value]
            elif isinstance(value, dict):
                _sanitize_leaves(value)

    def _update(style: dict, current: dict):
        for key, value in style.items():
            if key not in current:
                current[key] = value
                continue

            # Style leaves are by definition not to be recurred into
            if isinstance(value, dict) and (key not in style_leaves):
                _update(value, current[key])
            elif value is None:
                del current[key]
            else:
                current[key] = value

    old_style = copy_with_deep_values(current)

    try:
        if style is None:
            styles = []
        elif isinstance(style, (dict, str)):
            styles = [style]
        else:
            styles = list(style)

        if kwargs:
            styles.append(kwargs)

        for style in styles:
            if style == "default":
                reset()
            else:
                if isinstance(style, str):
                    current = get_style(style)
                else:
                    _sanitize_leaves(style)
                    unflatten_style(style)
                    _update(style, current)
    except:
        current = old_style
        raise


def reset() -> None:
    """Reset to default style."""
    global current
    current = copy_with_deep_values(styles["default"])


@contextmanager
def context(style: Optional[str | dict | Sequence] = None, **kwargs):
    """Create a style context for iplotx.

    Parameters:
        style: A single style specification or a list of style specifications, which are then
            applied in order. Each style can be a string (for an existing style) or a dictionary
            with the elements that are to change.
        **kwargs: Additional style changes to be applied at the end of all styles.

    Yields:
        A context manager that applies the style and reverts it back to the previous one upon exit.
    """
    current = get_style()
    try:
        use(style, **kwargs)
        yield
    finally:
        use(["default", current])


def unflatten_style(
    style_flat: dict[str, str | dict | int | float],
) -> None:
    """Convert a flat or semi-flat style into a fully structured dict.

    Parameters:
        style_flat: A flat dictionary where keys may contain underscores, which are taken to signify
            subdictionaries.

    NOTE: The dict is changed *in place*.

    Example:
        >>> style = {'vertex_size': 20}
        >>> unflatten_style(style)
        >>> print(style)
        {'vertex': {'size': 20}}
    """

    def _inner(style_flat: dict):
        keys = list(style_flat.keys())

        for key in keys:
            if "_" not in key:
                continue

            keyhead, keytail = key.split("_", 1)
            value = style_flat.pop(key)
            if keyhead not in style_flat:
                style_flat[keyhead] = {
                    keytail: value,
                }
            else:
                style_flat[keyhead][keytail] = value

        for key, value in style_flat.items():
            if isinstance(value, dict) and (key not in style_leaves):
                _inner(value)

    # top-level adjustments
    if "zorder" in style_flat:
        style_flat["network_zorder"] = style_flat["grouping_zorder"] = style_flat.pop(
            "zorder"
        )

    # Begin recursion
    _inner(style_flat)


def rotate_style(
    style: dict[str, Any],
    index: Optional[int] = None,
    key: Optional[Hashable] = None,
    props: Optional[Sequence[str]] = None,
) -> dict[str, Any]:
    """Rotate leaves of a style for a certain index or key.

    Parameters:
        style: The style to rotate.
        index: The integer to rotate the style leaves into.
        key: For dict-like leaves (e.g. vertex properties specified as a dict-like object over the
            vertices themselves), the key to use for rotation (e.g. the vertex itself).
        props: The properties to rotate, usually all leaf properties.

    Returns:
        A style with rotated leaves, which describes the properties of a single element (e.g.
        vertex).

    Example:
        >>> style = {'vertex': {'size': [10, 20]}}
        >>> rotate_style(style, index=0)
        {'vertex': {'size': 10}}
    """
    if (index is None) and (key is None):
        raise ValueError(
            "At least one of 'index' or 'key' must be provided to rotate_style."
        )

    if props is None:
        props = tuple(prop for prop in style_leaves if prop not in nonrotating_leaves)

    style = copy_with_deep_values(style)

    for prop in props:
        val = style.get(prop, None)
        if val is None:
            continue
        # Try integer indexing for ordered types
        if (index is not None) and isinstance(
            val, (tuple, list, np.ndarray, pd.Index, pd.Series)
        ):
            style[prop] = np.asarray(val)[index % len(val)]
        # Try key indexing for unordered, dict-like types
        if (
            (key is not None)
            and (not isinstance(val, (str, tuple, list, np.ndarray)))
            and hasattr(val, "__getitem__")
        ):
            # If only a subset of keys is provided, default the other ones
            # to the empty type constructor (e.g. 0 for ints, 0.0 for floats,
            # empty strings).
            if key in val:
                style[prop] = val[key]
            else:
                valtype = type(next(iter(val.values())))
                style[prop] = valtype()

    return style



"""
Edge ports
==========

This example demonstrates how to use **ports**, which specify at which angle edges exit from and enter into vertices.

Port directions are single or double letters borrowed from geography: `n` (north or up), `s` (south or down), `e` (east or right), `w` (west or left). `nw` (north-west) and similar directions are also allowed.

Each edge has two ports, one for its exit from the source vertex, and one for its entry into the target vertex. For undirected graphs, these refer to the order in the graph data structure.

If only one port is needed, you can set the other one to `None`.
"""

import igraph as ig
import numpy as np
import matplotlib.pyplot as plt
import iplotx as ipx

g = ig.Graph.Ring(3, directed=True)

ipx.plot(
    g,
    layout="circle",
    edge_curved=True,
    edge_ports=[
        ("n", "w"),  # exit from the top, enter from the left
        ("e", "s"),  # exit from the right, enter from the bottom
        ("n", "s"),  # exit from the top, enter from the bottom
    ],
    edge_tension=[1.5, 1.8, 0.8],
    edge_color=["tomato", "steelblue", "purple"],
)

# %%
# Ports react to inversion of the x- or y-axis and set of axis limits from positive to negatives, e.g.:

fig, ax = plt.subplots()
ipx.plot(
    g,
    ax=ax,
    layout="circle",
    edge_curved=True,
    edge_ports=[
        ("n", "w"),  # exit from the top, enter from the left
        ("e", "s"),  # exit from the right, enter from the bottom
        ("n", "s"),  # exit from the top, enter from the bottom
    ],
    edge_tension=[1.5, 1.8, 0.8],
    edge_color=["tomato", "steelblue", "purple"],
)
ax.invert_xaxis()
ax.invert_yaxis()

# %%
# is the mirror image (in both x and y) of the previous plot. Here's the same example but setting some ports to None:

g = ig.Graph.Ring(3, directed=True)

ipx.plot(
    g,
    layout="circle",
    edge_curved=True,
    edge_ports=[
        (None, "w"),
        ("e", "s"),
        ("n", None),
    ],
    edge_tension=[1.0, 1.8, 0.5],
)



from typing import (
    Any,
    Optional,
    Sequence,
)
from functools import partialmethod

from ...typing import (
    TreeDataProvider,
)


class Ete4DataProvider(TreeDataProvider):
    def _traverse(self, order: str) -> Any:
        """Traverse the tree."""
        return self.tree.traverse(order)

    preorder = partialmethod(_traverse, order="preorder")
    postorder = partialmethod(_traverse, order="postorder")

    def get_leaves(self) -> Sequence[Any]:
        return self.tree.leaves()

    @staticmethod
    def get_children(node: Any) -> Sequence[Any]:
        return node.children

    @staticmethod
    def get_branch_length(node: Any) -> Optional[float]:
        return node.dist

    @staticmethod
    def check_dependencies() -> bool:
        try:
            from ete4 import Tree
        except ImportError:
            return False
        return True

    @staticmethod
    def tree_type():
        from ete4 import Tree

        return Tree



"""
Chess masters
=============

This example from networkx shows the relatinoships between chess masters. The example has a lot of network analysis code that is somewhat irrelevant for `iplotx` and is kept to demonstrate functional parity.

.. note::
  The layout of this graph is computed with the Kamada-Kawai algorithm. In networkx, that requires the `scipy` package to be installed. Here, we also demonstrate the alternative, which is to convert the network to `igraph`, which has an internal implementation of the algorithm. While the exact layout is not identical, it is equally accurate and visually similar.
"""

import numpy as np
import networkx as nx
import igraph as ig
import matplotlib.pyplot as plt
import iplotx as ipx


# tag names specifying what game info should be
# stored in the dict on each digraph edge
game_details = ["Event", "Date", "Result", "ECO", "Site"]


def chess_pgn_graph(pgn_file="chess_masters_WCC.pgn.bz2"):
    """Read chess games in pgn format in pgn_file.

    Filenames ending in .bz2 will be uncompressed.

    Return the MultiDiGraph of players connected by a chess game.
    Edges contain game data in a dict.

    """
    import bz2

    G = nx.MultiDiGraph()
    game = {}
    with bz2.BZ2File(pgn_file) as datafile:
        lines = [line.decode().rstrip("\r\n") for line in datafile]
    for line in lines:
        if line.startswith("["):
            tag, value = line[1:-1].split(" ", 1)
            game[str(tag)] = value.strip('"')
        else:
            # empty line after tag set indicates
            # we finished reading game info
            if game:
                white = game.pop("White")
                black = game.pop("Black")
                G.add_edge(white, black, **game)
                game = {}
    return G


G = chess_pgn_graph()

print(
    f"Loaded {G.number_of_edges()} chess games between {G.number_of_nodes()} players\n"
)

# identify connected components of the undirected version
H = G.to_undirected()
Gcc = [H.subgraph(c) for c in nx.connected_components(H)]
if len(Gcc) > 1:
    print(f"Note the disconnected component consisting of:\n{Gcc[1].nodes()}")

# find all games with B97 opening (as described in ECO)
openings = {game_info["ECO"] for (white, black, game_info) in G.edges(data=True)}
print(f"\nFrom a total of {len(openings)} different openings,")
print("the following games used the Sicilian opening")
print('with the Najdorff 7...Qb6 "Poisoned Pawn" variation.\n')

for white, black, game_info in G.edges(data=True):
    if game_info["ECO"] == "B97":
        summary = f"{white} vs {black}\n"
        for k, v in game_info.items():
            summary += f"   {k}: {v}\n"
        summary += "\n"
        print(summary)

# make new undirected graph H without multi-edges
H = nx.Graph(G)

# edge width is proportional number of games played
edgewidth = [len(G.get_edge_data(u, v)) for u, v in H.edges()]

# node size is proportional to number of games won
wins = dict.fromkeys(G.nodes(), 0.0)
for u, v, d in G.edges(data=True):
    r = d["Result"].split("-")
    if r[0] == "1":
        wins[u] += 1.0
    elif r[0] == "1/2":
        wins[u] += 0.5
        wins[v] += 0.5
    else:
        wins[v] += 1.0
nodesize = [wins[v] * 1.5 for v in H]

# Generate layout for visualization
try:
    pos = nx.kamada_kawai_layout(H)
    ylegend = 0.1
except ImportError:
    Hig = ig.Graph.from_networkx(H)
    layout = Hig.layout("kk", weights=edgewidth).coords
    pos = {x: np.asarray(row) for x, row in zip(Hig.vs["_nx_name"], layout)}
    ylegend = 0.89

fig, ax = plt.subplots(figsize=(12, 12))
ipx.plot(
    H,
    layout=pos,
    ax=ax,
    title="World Chess Championship Games: 1886 - 1985",
    vertex_labels=True,
    style={
        "vertex": {
            "facecolor": "#210070",
            "edgecolor": "none",
            "alpha": 0.9,
            "size": nodesize,
            "label": {
                "color": "black",
                "size": 14,
                "bbox": {
                    "facecolor": "white",
                    "edgecolor": "black",
                    "alpha": 0.7,
                },
            },
        },
        "edge": {
            "alpha": 0.3,
            "linewidth": edgewidth,
            "color": "m",
        },
    },
)

ax.text(
    0.80,
    ylegend,
    "edge width = # games played\nnode size = # games won",
    size=14,
    horizontalalignment="center",
    transform=ax.transAxes,
    color="red",
)

ax.margins(0.1, 0.05)
fig.tight_layout()



"""
Knuth Miles US cities
=====================

This example from newtorkx shows an undirected graph of 128 US cities.
"""

import gzip
import re

import numpy as np
import matplotlib.pyplot as plt
import networkx as nx
import iplotx as ipx

# Ignore any warnings related to downloading shpfiles with cartopy
import warnings

warnings.simplefilter("ignore")


def miles_graph():
    """Return the cites example graph in miles_dat.txt
    from the Stanford GraphBase.
    """
    # open file miles_dat.txt.gz (or miles_dat.txt)

    fh = gzip.open("knuth_miles.txt.gz", "r")

    G = nx.Graph()
    G.position = {}
    G.population = {}

    cities = []
    for line in fh.readlines():
        line = line.decode()
        if line.startswith("*"):  # skip comments
            continue

        numfind = re.compile(r"^\d+")

        if numfind.match(line):  # this line is distances
            dist = line.split()
            for d in dist:
                G.add_edge(city, cities[i], weight=int(d))
                i = i + 1
        else:  # this line is a city, position, population
            i = 1
            (city, coordpop) = line.split("[")
            cities.insert(0, city)
            (coord, pop) = coordpop.split("]")
            (y, x) = coord.split(",")

            G.add_node(city)
            # assign position - Convert string to lat/long
            G.position[city] = (-float(x) / 100, float(y) / 100)
            G.population[city] = float(pop) / 1000
    return G


G = miles_graph()

print("Loaded miles_dat.txt containing 128 cities.")
print(G)

# make new graph of cites, edge if less than 300 miles between them
H = nx.Graph()
for v in G:
    H.add_node(v)
for u, v, d in G.edges(data=True):
    if d["weight"] < 300:
        H.add_edge(u, v)

# nodes colored by degree sized by population
node_color = [float(H.degree(v)) for v in H]

# plot with iplotx
fig, ax = plt.subplots(figsize=(10, 7))
ipx.plot(
    H,
    layout=G.position,
    ax=ax,
    style={
        "vertex": {
            "size": [2 * G.population[v] ** 0.5 for v in H],
            "facecolor": node_color,
            "cmap": "viridis",
        }
    },
)

# %%
# If you compare the style settings with the networkx example, you notice that
# `iplotx` sets the vertex diameter proportional with the "size" parameters,
# while `networkx` follows the `pylab.scatter` convention of scaling the diameter
# of each dot with the **square root** of the chosen size. Therefore, to
# recover the same plot as in the networkx example, we need to scale the
# size manually with the square root of the population.
#
# If we scale the vertex diameter with the population, it would look like this:

fig, ax = plt.subplots(figsize=(10, 7))
ipx.plot(
    H,
    layout=G.position,
    ax=ax,
    style={
        "vertex": {
            "size": [0.1 * G.population[v] for v in H],
            "facecolor": node_color,
            "cmap": "viridis",
        }
    },
)

# %%
# As you can see, the plot is almost identical but larger cities are more prominent.
#
# .. note::
#   There is no right or wrong behaviour here. In fact, not even `matplotlib` is internally consistent
#   with this, and `seaborn` and `matplotlib` use different conventions. If you want to have more dynamic range
#   (i.e. see smaller dots more clearly), scaling with the square root like in `networkx`, or even with the
#   cubic root or logarithm of the size is a good idea. If you want the user to imagine diameters/linear sizes
#   as related to the population, you can use a linear scaling, which is the default in `iplotx`.
#
# If you are curious about why `plotx` scales linearly with the diameter, the explanation is quite simple.
# `networkx` and others take the square root under the assumption that in that way the area of the node marker
# is proportional to the chosen size. However, that works less well for certain shapes, such as crosses, X, or
# lines. In that (not entirely uncommon) case, there is no area as such and the scaling as the square root ends up
# becoming a headache for everyone. This is no criticism of `networkx`, which is great. Just a little backstory.



# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html
import sys
import os
from pathlib import Path
from packaging.version import parse as parse_version

import iplotx

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

project = "iplotx"
copyright = "2025-%Y, Fabio Zanini"
author = "Fabio Zanini"
release = iplotx.__version__

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.napoleon",
    "myst_parser",
    "sphinx_design",
    "sphinx_gallery.gen_gallery",
]

# Napoleon settings
napoleon_google_docstring = True
napoleon_numpy_docstring = False
napoleon_include_init_with_doc = False
napoleon_include_private_with_doc = False
napoleon_include_special_with_doc = True
napoleon_use_admonition_for_examples = False
napoleon_use_admonition_for_notes = False
napoleon_use_admonition_for_references = False
napoleon_use_ivar = False
napoleon_use_param = True
napoleon_use_rtype = True
napoleon_preprocess_types = False
napoleon_type_aliases = None
napoleon_attr_annotations = True

myst_enable_extensions = ["colon_fence"]

sphinx_gallery_conf = {
    "examples_dirs": "../../gallery",  # path to your example scripts
    "gallery_dirs": "./gallery",  # path to where to save gallery generated output
    #'matplotlib_animations': (True, 'mp4'),
    "matplotlib_animations": True,
}

source_suffix = {
    ".rst": "restructuredtext",
    ".md": "markdown",
    ".myst": "markdown",
    ".txt": "markdown",
}


templates_path = ["_templates"]
exclude_patterns = []


# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_theme = "sphinx_rtd_theme"
html_static_path = ["_static"]

# -----------------------------------------------------------------------------
# Source code links (credit to the matplotlib project for this part)
# -----------------------------------------------------------------------------
link_github = True
# You can add build old with link_github = False

if link_github:
    import inspect

    extensions.append("sphinx.ext.linkcode")

    def linkcode_resolve(domain, info):
        """
        Determine the URL corresponding to Python object
        """
        if domain != "py":
            return None

        modname = info["module"]
        fullname = info["fullname"]

        submod = sys.modules.get(modname)
        if submod is None:
            return None

        obj = submod
        for part in fullname.split("."):
            try:
                obj = getattr(obj, part)
            except AttributeError:
                return None

        if inspect.isfunction(obj):
            obj = inspect.unwrap(obj)
        try:
            fn = inspect.getsourcefile(obj)
        except TypeError:
            fn = None
        if not fn or fn.endswith("__init__.py"):
            try:
                fn = inspect.getsourcefile(sys.modules[obj.__module__])
            except (TypeError, AttributeError, KeyError):
                fn = None
        if not fn:
            return None

        try:
            source, lineno = inspect.getsourcelines(obj)
        except (OSError, TypeError):
            lineno = None

        linespec = f"#L{lineno:d}-L{lineno + len(source) - 1:d}" if lineno else ""

        startdir = Path(iplotx.__file__).parent.parent
        try:
            fn = os.path.relpath(fn, start=startdir).replace(os.path.sep, "/")
        except ValueError:
            return None

        if not fn.startswith("iplotx/"):
            return None

        version = parse_version(iplotx.__version__)
        tag = "main" if version.is_devrelease else f"v{version.public}"
        return "https://github.com/fabilab/iplotx/blob" f"/{tag}/{fn}{linespec}"

else:
    extensions.append("sphinx.ext.viewcode")



"""
Biopython tree
==============

This example shows how to use `iplotx` to plot trees from Biopython's Phylo module.
"""

from Bio import Phylo
from io import StringIO
import matplotlib.pyplot as plt
import iplotx as ipx

# Make a tree from a string in Newick format
tree = next(
    Phylo.NewickIO.parse(
        StringIO(
            "(()(()((()())(()()))))",
        )
    )
)

ipx.plotting.tree(
    tree,
)

# %%
# To add labels to the leaves, you can use the `vertex_labels` argument as a dictionary.
# Moreover, the plot can be customised further using hmargin (horizontal label margin)
# and some `matplotlib` settings.

# sphinx_gallery_thumbnail_number = 2
leaf_labels = {leaf: f"Species {i+1}" for i, leaf in enumerate(tree.get_terminals())}
fig, ax = plt.subplots(figsize=(4, 4))
ipx.plotting.tree(
    tree,
    ax=ax,
    vertex_labels=leaf_labels,
    vertex_label_hmargin=10,
)
ax.set_xlim(0, 8)
ax.invert_yaxis()

# %%
# `iplotx` can compute a radial tree layout as well, and usual style modifications
# work for trees same as networks:

ipx.plotting.tree(
    tree,
    layout="radial",
    style=[
        "tree",
        {
            "edge": {
                "color": ["black", "pink"],
                "linewidth": 4,
            },
        },
    ],
    aspect=1,
    title="Black pink in your area",
)



"""
Module containing code to manipulate vertex visualisations, especially the VertexCollection class.
"""

from typing import (
    Optional,
    Sequence,
    Any,
)
import warnings
import numpy as np
import pandas as pd
import matplotlib as mpl
from matplotlib.collections import PatchCollection
from matplotlib.patches import (
    Patch,
    Ellipse,
    Circle,
    RegularPolygon,
    Rectangle,
)

from .style import (
    get_style,
    rotate_style,
    copy_with_deep_values,
)
from .utils.matplotlib import (
    _get_label_width_height,
    _build_cmap_fun,
    _forwarder,
)
from .label import LabelCollection


@_forwarder(
    (
        "set_clip_path",
        "set_clip_box",
        "set_snap",
        "set_sketch_params",
        "set_animated",
        "set_picker",
    )
)
class VertexCollection(PatchCollection):
    """Collection of vertex patches for plotting."""

    _factor = 1.0

    def __init__(
        self,
        layout: pd.DataFrame,
        *args,
        layout_coordinate_system: str = "cartesian",
        style: Optional[dict[str, Any]] = None,
        labels: Optional[Sequence[str]] = None,
        **kwargs,
    ):
        """Initialise the VertexCollection.

        Parameters:
            layout: The vertex layout.
            layout_coordinate_system: The coordinate system for the layout, usually "cartesian").
            style: The vertex style (subdictionary "vertex") to apply.
            labels: The vertex labels, if present.
        """

        self._index = layout.index
        self._style = style
        self._labels = labels
        self._layout = layout
        self._layout_coordinate_system = layout_coordinate_system

        # Create patches from structured data
        patches, sizes, kwargs2 = self._init_vertex_patches()

        kwargs.update(kwargs2)
        kwargs["match_original"] = True

        # Pass to PatchCollection constructor
        super().__init__(patches, *args, **kwargs)

        # Set offsets in coordinate system
        self._update_offsets_from_layout()

        # Compute _transforms like in _CollectionWithScales for dpi issues
        self.set_sizes(sizes)

        if self._labels is not None:
            self._compute_label_collection()

    def get_children(self) -> tuple[mpl.artist.Artist]:
        """Get the children artists.

        This can include the labels as a LabelCollection.
        """
        children = []
        if hasattr(self, "_label_collection"):
            children.append(self._label_collection)
        return tuple(children)

    def set_figure(self, fig) -> None:
        """Set the figure for this artist and all children."""
        super().set_figure(fig)
        self.set_sizes(self._sizes, self.get_figure(root=True).dpi)
        self._update_children()
        for child in self.get_children():
            child.set_figure(fig)

    def get_index(self):
        """Get the VertexCollection index."""
        return self._index

    def get_vertex_id(self, index):
        """Get the id of a single vertex at a positional index."""
        return self._index[index]

    def get_sizes(self):
        """Get vertex sizes (max of width and height), not scaled by dpi."""
        return self._sizes

    def get_sizes_dpi(self):
        """Get vertex sizes (max of width and height), scaled by dpi."""
        return self._transforms[:, 0, 0]

    def set_sizes(self, sizes, dpi: float = 72.0) -> None:
        """Set vertex sizes.

        This rescales the current vertex symbol/path linearly, using this
        value as the largest of width and height.

        @param sizes: A sequence of vertex sizes or a single size.
        """
        if sizes is None:
            self._sizes = np.array([])
            self._transforms = np.empty((0, 3, 3))
        else:
            self._sizes = np.asarray(sizes)
            self._transforms = np.zeros((len(self._sizes), 3, 3))
            scale = self._sizes * dpi / 72.0 * self._factor
            self._transforms[:, 0, 0] = scale
            self._transforms[:, 1, 1] = scale
            self._transforms[:, 2, 2] = 1.0
        self.stale = True

    get_size = get_sizes
    set_size = set_sizes

    def get_layout(self) -> pd.DataFrame:
        """Get the vertex layout.

        Returns:
            The vertex layout as a DataFrame.
        """
        return self._layout

    def get_layout_coordinate_system(self) -> str:
        """Get the layout coordinate system.

        Returns:
            Name of the layout coordinate system, e.g. "cartesian" or "polar".
        """
        return self._layout_coordinate_system

    def get_offsets(self, ignore_layout: bool = True) -> np.ndarray:
        """Get the vertex offsets.

        Parameters:
            ignore_layout: If True, return the matplotlib Artist._offsets directly, ignoring the
                layout coordinate system. If False, it's equivalent to get_layout().values.

        Returns:
            The vertex offsets as a 2D numpy array.

        Note: It is best for users to *not* ignore the layout coordinate system, as it may lead
        to inconsistencies. However, some internal matplotlib functions require the default
        signature of this function to look at the vanilla offsets, hence the default parameters.
        """
        if not ignore_layout:
            return self.get_layout().values
        else:
            return self._offsets

    def _update_offsets_from_layout(self) -> None:
        """Update offsets in matplotlib coordinates from the layout DataFrame."""
        if self._layout_coordinate_system == "cartesian":
            self._offsets = self._layout.values
        elif self._layout_coordinate_system == "polar":
            # Convert polar coordinates (r, theta) to cartesian (x, y)
            r = self._layout.iloc[:, 0].values
            theta = self._layout.iloc[:, 1].values
            if self._offsets is None:
                self._offsets = np.zeros((len(r), 2))
            self._offsets[:, 0] = r * np.cos(theta)
            self._offsets[:, 1] = r * np.sin(theta)
        else:
            raise ValueError(
                f"Layout coordinate system not supported: {self._layout_coordinate_system}."
            )

    def set_offsets(self, offsets: np.ndarray) -> None:
        """Set the vertex positions/offsets in layout coordinates.

        Parameters:
            offsets: Array of coordinates in the layout coordinate system. For polar layouts,
                these should be in the form of (r, theta) pairs.
        """
        self._layout.values[:] = offsets
        self._update_offsets_from_layout()
        self.stale = True

    def get_style(self) -> Optional[dict[str, Any]]:
        """Get the style dictionary for the vertices."""
        return self._style

    def _init_vertex_patches(self):
        style = self._style or {}
        if "cmap" in style:
            cmap_fun = _build_cmap_fun(
                style["facecolor"],
                style["cmap"],
            )
        else:
            cmap_fun = None

        if style.get("size", 20) == "label":
            if self._labels is None:
                warnings.warn(
                    "No labels found, cannot resize vertices based on labels."
                )
                style["size"] = get_style("default.vertex")["size"]

        if "cmap" in style:
            colorarray = []
        patches = []
        sizes = []
        for i, (vid, row) in enumerate(self._layout.iterrows()):
            stylei = rotate_style(style, index=i, key=vid)
            if stylei.get("size", 20) == "label":
                stylei["size"] = _get_label_width_height(
                    str(self._labels[vid]), **style.get("label", {})
                )
            if cmap_fun is not None:
                colorarray.append(style["facecolor"])
                stylei["facecolor"] = cmap_fun(stylei["facecolor"])

            # Shape of the vertex (Patch)
            art, size = make_patch(**stylei)
            patches.append(art)
            sizes.append(size)

        kwargs = {}
        if "cmap" in style:
            vmin = np.min(colorarray)
            vmax = np.max(colorarray)
            norm = mpl.colors.Normalize(vmin=vmin, vmax=vmax)
            kwargs["cmap"] = style["cmap"]
            kwargs["norm"] = norm

        return patches, sizes, kwargs

    def _compute_label_collection(self):
        transform = self.get_offset_transform()

        style = (
            copy_with_deep_values(self._style.get("label", None))
            if self._style is not None
            else {}
        )
        forbidden_props = ["hpadding", "vpadding"]
        for prop in forbidden_props:
            if prop in style:
                del style[prop]

        self._label_collection = LabelCollection(
            self._labels,
            style=style,
            offsets=self._offsets,
            transform=transform,
        )

    def get_labels(self):
        """Get the vertex labels.

        Returns:
            The artist with the LabelCollection.
        """

        if hasattr(self, "_label_collection"):
            return self._label_collection
        else:
            return None

    @property
    def stale(self):
        return super().stale

    @stale.setter
    def stale(self, val):
        PatchCollection.stale.fset(self, val)
        if val and hasattr(self, "stale_callback_post"):
            self.stale_callback_post(self)

    def _update_children(self) -> None:
        """Update children before drawing and before first render."""
        self._update_labels()

    def _update_labels(self) -> None:
        """Update labels before drawing.

        NOTE: This needs to work in figure coordinates.
        """
        if not hasattr(self, "_label_collection"):
            return

        if self.get_layout_coordinate_system() != "polar":
            return

        transform = self.get_offset_transform()
        trans = transform.transform

        zero_fig = trans(np.array([0, 0]))
        offsets_fig = trans(self.get_labels().get_offsets())
        doffsets_fig = offsets_fig - zero_fig
        rotations = np.arctan2(doffsets_fig[:, 1], doffsets_fig[:, 0])
        self.get_labels().set_rotations(rotations)

    @mpl.artist.allow_rasterization
    def draw(self, renderer):
        if not self.get_visible():
            return

        # null graph, no need to draw anything
        # NOTE: I would expect this to be already a clause in the superclass by oh well
        if len(self.get_paths()) == 0:
            return

        self.set_sizes(self._sizes, self.get_figure(root=True).dpi)

        # Set the label rotations already, hopefully this is not too early
        self._update_children()

        # NOTE: This draws the vertices first, then the labels.
        # The correct order would be vertex1->label1->vertex2->label2, etc.
        # We might fix if we manage to find a way to do it.
        super().draw(renderer)
        for child in self.get_children():
            child.draw(renderer)


def make_patch(
    marker: str = "o",
    size: float | Sequence[float] = 20,
    **kwargs,
) -> tuple[Patch, float]:
    """Make a patch of the given marker shape and size."""
    forbidden_props = ["label", "cmap", "norm", "cascade"]
    for prop in forbidden_props:
        if prop in kwargs:
            kwargs.pop(prop)

    if np.isscalar(size):
        size = float(size)
        size = (size, size)

    # Size of vertices is determined in self._transforms, which scales with dpi, rather than here,
    # so normalise by the average dimension (btw x and y) to keep the ratio of the marker.
    # If you check in get_sizes, you will see that rescaling also happens with the max of width
    # and height.
    size = np.asarray(size, dtype=float)
    size_max = size.max()
    if size_max > 0:
        size /= size_max

    art: Patch
    if marker in ("o", "c", "circle"):
        art = Circle((0, 0), size[0] / 2, **kwargs)
    elif marker in ("s", "square", "r", "rectangle"):
        art = Rectangle((-size[0] / 2, -size[1] / 2), size[0], size[1], **kwargs)
    elif marker in ("^", "triangle"):
        art = RegularPolygon((0, 0), numVertices=3, radius=size[0] / 2, **kwargs)
    elif marker in ("d", "diamond"):
        art, _ = make_patch("s", size[0], angle=45, **kwargs)
    elif marker in ("v", "triangle_down"):
        art = RegularPolygon(
            (0, 0), numVertices=3, radius=size[0] / 2, orientation=np.pi, **kwargs
        )
    elif marker in ("e", "ellipse"):
        art = Ellipse((0, 0), size[0] / 2, size[1] / 2, **kwargs)
    else:
        raise KeyError(f"Unknown marker: {marker}")

    return (art, size_max)



"""
Tree cascades
=============

This example shows how to use `iplotx` to add cascading backgrounds to trees.
"Cascading" here means that each patch (rectangle/wedge/etc.) will cover a node
and all descendants, down to the leaves.
"""

from Bio import Phylo
from io import StringIO
import matplotlib.pyplot as plt
import iplotx as ipx

# Make a tree from a string in Newick format
tree = next(
    Phylo.NewickIO.parse(
        StringIO(
            "(()(()((()())(()()))))",
        )
    )
)

backgrounds = {
    tree.get_nonterminals()[3]: "turquoise",
    tree.get_terminals()[0]: "tomato",
    tree.get_terminals()[1]: "purple",
}

ipx.plotting.tree(
    tree,
    vertex_cascade_facecolor=backgrounds,
)

# %%
# Cascading patches have a style option "extend" which affects whether the patches extend to the end of the deepest leaf:

ipx.plotting.tree(
    tree,
    layout="vertical",
    vertex_cascade_facecolor=backgrounds,
    vertex_cascade_extend=True,
)

# %%
# Cascading patches work with radial layouts as well:

# sphinx_gallery_thumbnail_number = 3
ipx.plotting.tree(
    tree,
    layout="radial",
    vertex_cascade_facecolor=backgrounds,
    vertex_cascade_extend=True,
    aspect=1,
)



from typing import Optional, Sequence
from contextlib import nullcontext
import numpy as np
import pandas as pd
import matplotlib as mpl
import matplotlib.pyplot as plt

from .typing import (
    GraphType,
    LayoutType,
    GroupingType,
    TreeType,
)
from .network import NetworkArtist
from .groups import GroupingArtist
from .tree import TreeArtist
from .style import context


def network(
    network: Optional[GraphType] = None,
    layout: Optional[LayoutType] = None,
    grouping: Optional[GroupingType] = None,
    vertex_labels: Optional[list | dict | pd.Series] = None,
    edge_labels: Optional[Sequence] = None,
    ax: Optional[mpl.axes.Axes] = None,
    style: str | dict | Sequence[str | dict] = (),
    title: Optional[str] = None,
    aspect: Optional[str | float] = None,
    margins: float | tuple[float, float] = 0,
    **kwargs,
) -> list[mpl.artist.Artist]:
    """Plot this network and/or vertex grouping using the specified layout.

    Parameters:
        network: The network to plot. Can be a networkx or igraph graph.
        layout: The layout to use for plotting. If None, a layout will be looked for in the
            network object and, if none is found, an exception is raised. Defaults to None.
        vertex_labels: The labels for the vertices. If None or False, no vertex labels
            will be drawn. If a list, the labels are taken from the list. If a dict, the keys
            should be the vertex IDs and the values should be the labels. If True (a single
            bool value), the vertex IDs will be used as labels.
        edge_labels: The labels for the edges. If None, no edge labels will be drawn. Defaults
            to None.
        ax: The axis to plot on. If None, a new figure and axis will be created. Defaults to
            None.
        style: Apply this style for the objects to plot. This can be a sequence (e.g. list)
            of styles and they will be applied in order.
        title: If not None, set the axes title to this value.
        aspect: If not None, set the aspect ratio of the axis to this value. The most common
            value is 1.0, which proportionates x- and y-axes.
        margins: How much margin to leave around the plot. A higher value (e.g. 0.1) can be
            used as a quick fix when some vertex shapes reach beyond the plot edge. This is
            a fraction of the data limits, so 0.1 means 10% of the data limits will be left
            as margin.
        **kwargs: Additional arguments are treated as an alternate way to specify style. If
            both "style" and additional **kwargs are provided, they are both applied in that
            order (style, then **kwargs).

    Returns:
        A list of mpl.artist.Artist objects, set as a direct child of the matplotlib Axes.
        The list can have one or two elements, depending on whether you are requesting to
        plot a network, a grouping, or both.
    """
    stylecontext = context(style, **kwargs) if style or kwargs else nullcontext()

    with stylecontext:
        if (network is None) and (grouping is None):
            raise ValueError("At least one of network or grouping must be provided.")

        if ax is None:
            fig, ax = plt.subplots()

        artists = []
        if network is not None:
            nwkart = NetworkArtist(
                network,
                layout,
                vertex_labels=vertex_labels,
                edge_labels=edge_labels,
                transform=mpl.transforms.IdentityTransform(),
                offset_transform=ax.transData,
            )
            ax.add_artist(nwkart)

            # Set the figure, which itself sets the dpi scale for vertices, edges,
            # arrows, etc. Now data limits can be computed correctly
            nwkart.set_figure(ax.figure)

            artists.append(nwkart)

            # Set normailsed layout since we have it by now
            layout = nwkart.get_layout()

        if grouping is not None:
            grpart = GroupingArtist(
                grouping,
                layout,
                network=network,
                transform=ax.transData,
            )
            ax.add_artist(grpart)

            grpart.set_figure(ax.figure)
            artists.append(grpart)

        if title is not None:
            ax.set_title(title)

        if aspect is not None:
            ax.set_aspect(aspect)

        _postprocess_axis(ax, artists)

        if np.isscalar(margins):
            margins = (margins, margins)
        if (margins[0] != 0) or (margins[1] != 0):
            ax.margins(*margins)

        return artists


def tree(
    tree: Optional[TreeType] = None,
    layout: str | LayoutType = "horizontal",
    orientation: Optional[str] = None,
    directed: bool | str = False,
    vertex_labels: Optional[list | dict | pd.Series] = None,
    leaf_labels: Optional[list | dict | pd.Series] = None,
    ax: Optional[mpl.axes.Axes] = None,
    style: str | dict | Sequence[str | dict] = "tree",
    title: Optional[str] = None,
    aspect: Optional[str | float] = None,
    margins: float | tuple[float, float] = 0,
    **kwargs,
) -> TreeArtist:
    """Plot a tree using the specified layout.

    Parameters:
        tree: The tree to plot. Can be a BioPython.Phylo.Tree object.
        layout: The layout to use for plotting.
        orientation: The orientation of the layout. Can be "right" or "left". Defaults to
            "right" for horizontal layout, "descending" or "ascending" for vertical layout,
            and "clockwise" or "anticlockwise" for radial layout.
        directed: If False, donot draw arrows. If True or "child", draw arrows from parent to child
            node. If "parent", draw arrows the other way around.

    Returns:
        A TreeArtist object, set as a direct child of the matplotlib Axes.
    """
    stylecontext = context(style, **kwargs) if style or kwargs else nullcontext()

    with stylecontext:
        if ax is None:
            fig, ax = plt.subplots()

        artist = TreeArtist(
            tree=tree,
            layout=layout,
            orientation=orientation,
            directed=directed,
            transform=mpl.transforms.IdentityTransform(),
            offset_transform=ax.transData,
            vertex_labels=vertex_labels,
            leaf_labels=leaf_labels,
        )
        ax.add_artist(artist)

        artist.set_figure(ax.figure)

        if title is not None:
            ax.set_title(title)

        if aspect is not None:
            ax.set_aspect(aspect)

        _postprocess_axis(ax, [artist])

        if np.isscalar(margins):
            margins = (margins, margins)
        if (margins[0] != 0) or (margins[1] != 0):
            ax.margins(*margins)

    return artist


# INTERNAL ROUTINES
def _postprocess_axis(ax, artists):
    """Postprocess axis after plotting."""

    # Despine
    ax.spines["right"].set_visible(False)
    ax.spines["top"].set_visible(False)
    ax.spines["left"].set_visible(False)
    ax.spines["bottom"].set_visible(False)

    # Remove axis ticks
    ax.set_xticks([])
    ax.set_yticks([])

    # Set new data limits
    bboxes = []
    for art in artists:
        bboxes.append(art.get_datalim(ax.transData))
    bbox = mpl.transforms.Bbox.union(bboxes)
    ax.update_datalim(bbox)

    # Autoscale for x/y axis limits
    ax.autoscale_view()



"""
Typing hints for iplotx.

Some of these types are legit, others are just Any but renamed to improve readability throughout
the codebase.
"""

from typing import (
    Union,
    Sequence,
    Any,
    TypeVar,
)
from collections.abc import Hashable
import numpy as np
import pandas as pd


# NOTE: GraphType is supposed to indicate any kind of graph object that is accepted by
# iplotx's functions, e.g. igraph.Graph or networkx.Graph and subclasses. It is not
# quite possible to really statically type it because providers can add their own
# types - together with protocols to process them - at runtime.
# Nonetheless, for increased readibility we define separately-named types in this
# module to be used throughout the codebase.
GraphType = Any
TreeType = Any

# NOTE: The commented ones are not a mistake: they are supported but cannot be
# statically typed if the user has no igraph installed (it's a soft dependency).
LayoutType = Union[
    str,
    Sequence[Sequence[float]],
    np.ndarray,
    pd.DataFrame,
    # igraph.Layout,
]
GroupingType = Union[
    Sequence[set],
    Sequence[int],
    Sequence[str],
    # igraph.clustering.Clustering,
    # igraph.clustering.VertexClustering,
    # igraph.clustering.Cover,
    # igraph.clustering.VertexCover,
]


T = TypeVar("T")
LeafProperty = Union[
    T,
    Sequence[T],
    dict[Hashable, T],
]

Pair = tuple[T, T]



"""
Style specifications
====================

This minimal example shows how to style a plot in a few different ways.
"""

import networkx as nx
import matplotlib.pyplot as plt
import iplotx as ipx

# Create a graph and add a self-loop to node 0
G = nx.complete_graph(3, create_using=nx.DiGraph)
pos = nx.circular_layout(G)
ne = G.number_of_edges()

# Add self-loops to the remaining nodes
G.add_edge(0, 0)
edgelist = [(1, 1), (2, 2)]
G.add_edges_from(edgelist)

# Style the edge lines
linestyle = {e: "-" if e not in edgelist else "--" for e in G.edges()}

ipx.plot(
    G,
    layout=pos,
    vertex_labels=True,
    style={
        "vertex": {
            "size": 30,
            "facecolor": "lightblue",
            "edgecolor": "none",
            "label": {
                "color": "black",
            },
        },
        "edge": {
            "linestyle": linestyle,
            "paralleloffset": 0,
            "looptension": 3.5,
            "arrow": {
                "marker": "|>",
            },
        },
    },
)

# %%
# In addition to fully structured styles in the form of nested dictionaries, `iplotx`
# also accepts flat or semi-flat styles, with levels separated by underscores ("_").
# The above is equivalent to:

ipx.plot(
    G,
    layout=pos,
    vertex_labels=True,
    style=dict(
        vertex_size=30,
        vertex_facecolor="lightblue",
        vertex_edgecolor="none",
        vertex_label_color="black",
        edge_linestyle=linestyle,
        edge_paralleloffset=0,
        edge_looptension=3.5,
        edge_arrow_marker="|>",
    ),
)

# %%
# The `**kwargs` flexible argument of `ipx.plotting` can be used for this purpose as well,
# for increased readability:

ipx.plot(
    G,
    layout=pos,
    vertex_labels=True,
    vertex_size=30,
    vertex_facecolor="lightblue",
    vertex_edgecolor="none",
    vertex_label_color="black",
    edge_linestyle=linestyle,
    edge_paralleloffset=0,
    edge_looptension=3.5,
    edge_arrow_marker="|>",
)

# %%
# The same can be achieved also with a style **context**:

with ipx.style.context(
    vertex_size=30,
    vertex_facecolor="lightblue",
    vertex_edgecolor="none",
    vertex_label_color="black",
    edge_linestyle=linestyle,
    edge_paralleloffset=0,
    edge_looptension=3.5,
    edge_arrow_marker="|>",
):
    ipx.plot(
        G,
        layout=pos,
        vertex_labels=True,
    )
